"""Parsing of SPA actions."""
from nengo.exceptions import NetworkContextError, SpaParseError
from nengo.network import Network
from nengo_spa.modules.basalganglia import BasalGanglia
from nengo_spa.modules.thalamus import Thalamus
from nengo_spa.ast import (
    Action, ConstructionContext, DotProduct, Effect, Effects, Module,
    Reinterpret, Sink, Symbol, Translate)
from nengo.utils.compat import is_integer


class Parser(object):
    """Parser for SPA actions.

    Parameters
    ----------
    vocabs : dict
        Vocabularies to make accessible in the parsed action rules.
    """

    builtins = {
        'dot': DotProduct,
        'reinterpret': Reinterpret,
        'translate': Translate}

    def __init__(self, vocabs=None):
        if vocabs is None:
            vocabs = {}
        self.vocabs = vocabs

    def parse_action(self, action, index=0, name=None, strict=True):
        """Parse an SPA action.

        Parameters
        ----------
        action : str
            Action to parse.
        index : int, optional
            Index of the action for identification by basal ganglia and
            thalamus.
        name : str, optional
            Name of the action.
        strict : bool, optional
            If ``True`` only actual conditional actions are allowed and an
            exception will be raised for anything else. If ``False``, allows
            also the parsing of effects without the conditional part.

        Returns
        -------
        :class:`spa_ast.Action` or :class:`spa_ast.Effects`
        """
        try:
            condition, effects = action.split('-->', 1)
        except ValueError:
            if strict:
                raise SpaParseError("Not an action, '-->' missing.")
            else:
                return self.parse_effects(action, channeled=False)
        else:
            return Action(
                self.parse_expr(condition),
                self.parse_effects(effects, channeled=True), index, name=name)

    def parse_effects(self, effects, channeled=False):  # noqa: C901
        """Pares SPA effects.

        Parameters
        ----------
        effects : str
            Effects to parse.
        channeled : bool, optional
            Whether the effects should be passed through channels when
            constructed.

        Returns
        -------
        :class:`spa_ast.Effects`
        """
        parsed = []
        symbol_stack = []
        start = 0
        for i, c in enumerate(effects):
            top = symbol_stack[-1] if len(symbol_stack) > 0 else None
            if top == '\\':  # escaped character, ignore
                symbol_stack.pop()
            elif top is not None and top in '\'"':  # in a string
                if c == '\\':  # escape
                    symbol_stack.append(c)
                elif c == top:  # end string
                    symbol_stack.pop()
            else:
                if c in '\'"':  # start string
                    symbol_stack.append(c)
                elif c in '([':  # start parens/brackets
                    symbol_stack.append(c)
                elif c in ')]':  # end parens/brackets
                    if (top == '(' and c != ')') or (top == '[' and c != ']'):
                        raise SpaParseError("Parenthesis mismatch.")
                    symbol_stack.pop()
                elif c == ',' and len(symbol_stack) == 0:  # effect delimiter
                    parsed.append(effects[start:i])
                    start = i + 1
        parsed.append(effects[start:])

        if len(symbol_stack) != 0:
            top = symbol_stack.pop()
            if top in '([':
                raise SpaParseError("Parenthesis mismatch.")
            elif top in '\'"':
                raise SpaParseError("Unclosed string.")
            else:
                raise SpaParseError("Unmatched: " + top)

        return Effects(*[self.parse_effect(effect, channeled=channeled)
                         for effect in parsed])

    def parse_effect(self, effect, channeled=False):
        """Parse single SPA effect.

        Parameters
        ----------
        effect : str
            Effect to parse.
        channeled : bool, optional
            Whether the effect should be passed through a channel when
            constructed.

        Returns
        -------
        :class:`spa_ast.Effect`
        """
        try:
            sink, source = effect.split('=', 1)
        except ValueError:
            raise SpaParseError("Not an effect; assignment missing")
        return Effect(
            Sink(sink.strip()), self.parse_expr(source), channeled=channeled)

    def parse_expr(self, expr):
        """Parse an SPA expression.

        Parameters
        ----------
        expr : str
            Expression to parse.

        Returns
        -------
        :class:`spa_ast.Source`
        """
        return eval(expr, {}, self)

    def __getitem__(self, key):
        if key == '__tracebackhide__':  # gives better tracebacks in py.test
            return False
        if key in self.vocabs:
            return self.vocabs[key]
        if key in self.builtins:
            return self.builtins[key]
        if key[0].isupper():
            return Symbol(key)
        else:
            return Module(key)


class Actions(object):
    """A collection of Action objects.

    The *args and **kwargs are treated as unnamed and named Actions,
    respectively.

    The keyword argument `vocabs` is special in that it provides a dictionary
    mapping names to vocabularies. The vocabularies can then be used with those
    names in the action rules.
    """

    def __init__(self, *args, **kwargs):
        super(Actions, self).__init__()

        self.actions = []
        self.named_actions = {}

        self.construction_context = None

        self.parse(*args, **kwargs)

    def parse(self, *args, **kwargs):
        vocabs = kwargs.pop('vocabs', None)

        sorted_kwargs = sorted(kwargs.items())

        parser = Parser(vocabs=vocabs)
        for action in args:
            self._parse_and_add(parser, action)
        for name, action in sorted_kwargs:
            self._parse_and_add(parser, action, name=name)

    def __len__(self):
        return len(self.actions)

    def __getitem__(self, key):
        if is_integer(key):
            return self.actions[key]
        else:
            return self.named_actions[key]

    def _parse_and_add(self, parser, action, name=None):
        ast = parser.parse_action(
            action, len(self.actions), strict=False, name=name)
        self.actions.append(ast)
        if name is not None:
            self.named_actions[name] = ast

    def build(self, bg=None, thalamus=None):
        needs_bg = len(self.actions) > 0

        if len(Network.context) <= 0:
            raise NetworkContextError(
                "Actions.build can only be called inside a ``with network:`` "
                "block.")
        root_network = Network.context[-1]

        with root_network:
            if needs_bg and bg is None:
                bg = BasalGanglia(action_count=len(self.actions))
                root_network.bg = bg
            if needs_bg and thalamus is None:
                thalamus = Thalamus(action_count=len(self.actions))
                for i, a in enumerate(self.actions):
                    thalamus.actions.ensembles[i].label = (
                        'action[{}]: {}'.format(i, a.effects))
                thalamus.connect_bg(bg)
                root_network.thalamus = thalamus

        self.construction_context = ConstructionContext(
            root_network, bg=bg, thalamus=thalamus)
        with root_network:
            for action in self.actions:
                action.infer_types(root_network, None)
            # Infer types for all actions before doing any construction, so
            # that # all semantic pointers are added to the respective
            # vocabularies so that the translate transform are identical.
            for action in self.actions:
                action.construct(self.construction_context)

        return bg, thalamus, self.construction_context.constructed



# -*- coding: utf-8 -*-
#
# This file is execfile()d with the current directory set
# to its containing dir.

import os
import sys

try:
    import nengo_spa
    import sphinx_rtd_theme
except ImportError:
    print("To build the documentation, nengo_spa and sphinx_rtd_theme "
          "must be installed in the current environment. Please install these "
          "and their requirements first. A virtualenv is recommended!")
    sys.exit(1)

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.intersphinx',
    'sphinx.ext.mathjax',
    'sphinx.ext.todo',
    'sphinx.ext.viewcode',
    'numpydoc',
    'nengo.utils.docutils',
]

# -- sphinx.ext.autodoc
autoclass_content = 'both'  # class and __init__ docstrings are concatenated
autodoc_default_flags = ['members']
autodoc_member_order = 'bysource'  # default is alphabetical

# -- sphinx.ext.intersphinx
intersphinx_mapping = {
    'numpy': ('http://docs.scipy.org/doc/numpy', None)
}

# -- sphinx.ext.todo
todo_include_todos = True
# -- numpydoc config
numpydoc_show_class_members = False

# -- sphinx
exclude_patterns = ['_build']
source_suffix = '.rst'
source_encoding = 'utf-8'
master_doc = 'index'

# Need to include https Mathjax path for sphinx < v1.3
mathjax_path = ("https://cdn.mathjax.org/mathjax/latest/MathJax.js"
                "?config=TeX-AMS-MML_HTMLorMML")

project = u'Nengo SPA'
authors = u'Applied Brain Research'
copyright = nengo_spa.__copyright__
version = '.'.join(nengo_spa.__version__.split('.')[:2])  # Short X.Y version
release = nengo_spa.__version__  # Full version, with tags
pygments_style = 'default'

# -- Options for HTML output --------------------------------------------------

html_theme = 'sphinx_rtd_theme'
html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]
html_title = "Nengo {0} docs".format(release)
html_static_path = ['_static']
html_context = {
    'css_files': [os.path.join('_static', 'custom.css')],
}
html_use_smartypants = True
htmlhelp_basename = 'Nengodoc'
html_last_updated_fmt = ''  # Suppress 'Last updated on:' timestamp
html_show_sphinx = False

# -- Options for LaTeX output -------------------------------------------------

latex_elements = {
    'papersize': 'letterpaper',
    'pointsize': '11pt',
    # 'preamble': '',
}

latex_documents = [
    # (source start file, target, title, author, documentclass [howto/manual])
    ('index', 'nengo.tex', html_title, authors, 'manual'),
]

# -- Options for manual page output -------------------------------------------

man_pages = [
    # (source start file, name, description, authors, manual section).
    ('index', 'nengo', html_title, [authors], 1)
]

# -- Options for Texinfo output -----------------------------------------------

texinfo_documents = [
    # (source start file, target, title, author, dir menu entry,
    #  description, category)
    ('index', 'nengo', html_title, authors, 'Nengo',
     'Large-scale neural simulation in Python', 'Miscellaneous'),
]



import numpy as np
from numpy.testing import assert_allclose

import nengo
from nengo_spa.networks.selection import IA, Thresholding, WTA


def test_ia(Simulator, plt, seed):
    def input_fn(t):
        if t < 0.2:
            return [0.8, 0.5, 0.2, -0.1]
        else:
            return [0.1, -0.4, 0., 0.3]

    with nengo.Network(seed=seed) as m:
        ia = IA(50, 4)
        in_node = nengo.Node(input_fn)
        nengo.Connection(in_node, ia.input)

        reset_node = nengo.Node(lambda t: 0.4 < t < 0.5)
        nengo.Connection(reset_node, ia.input_reset, synapse=0.1)

        in_p = nengo.Probe(in_node)
        reset_p = nengo.Probe(reset_node)
        accum_p = nengo.Probe(ia.accumulators.output, synapse=0.01)
        out_p = nengo.Probe(ia.output, synapse=0.03)

    with Simulator(m) as sim:
        sim.run(0.9)
    t = sim.trange()

    plt.subplot(3, 1, 1)
    plt.plot(t, sim.data[in_p])
    plt.plot(t, sim.data[reset_p], c='k')
    plt.ylabel("Input")
    plt.subplot(3, 1, 2)
    plt.plot(t, sim.data[accum_p])
    plt.ylabel("Accumulators")
    plt.subplot(3, 1, 3)
    plt.plot(t, sim.data[out_p])
    plt.ylabel("Output")
    plt.xlabel("Time")

    first_selection = np.logical_and(0.15 < t, t < 0.4)
    assert np.all(sim.data[out_p][first_selection, 0] > 0.85)
    assert_allclose(sim.data[out_p][first_selection, 1:], 0., atol=0.05)
    assert np.all(sim.data[out_p][t > 0.85, 3] > 0.85)
    assert_allclose(sim.data[out_p][t > 0.85, :3], 0., atol=0.05)


def test_thresholding_array(Simulator, plt, seed):
    with nengo.Network(seed=seed) as m:
        thresholding = Thresholding(50, 4, 0.2, function=lambda x: x > 0.)
        in_node = nengo.Node([0.8, 0.5, 0.2, -0.1], label='input')
        nengo.Connection(in_node, thresholding.input)

        in_p = nengo.Probe(in_node)
        out_p = nengo.Probe(thresholding.output, synapse=0.03)
        thresholded_p = nengo.Probe(thresholding.thresholded, synapse=0.03)

    with Simulator(m) as sim:
        sim.run(0.2)
    t = sim.trange()

    plt.subplot(3, 1, 1)
    plt.plot(t, sim.data[in_p])
    plt.ylabel("Input")
    plt.subplot(3, 1, 2)
    plt.plot(t, sim.data[out_p])
    plt.plot(t[t > 0.15], 0.9 * np.ones(t.shape)[t > 0.15], c='k', lw=2)
    plt.ylabel("Output")
    plt.subplot(3, 1, 3)
    plt.plot(t, sim.data[thresholded_p])
    plt.plot(t[t > 0.15], 0.6 * np.ones(t.shape)[t > 0.15], c='k', lw=2)
    plt.plot(t[t > 0.15], 0.3 * np.ones(t.shape)[t > 0.15], c='k', lw=2)
    plt.ylabel("Utilities")

    assert np.all(sim.data[out_p][t > 0.15, :2] > 0.9)
    assert np.all(sim.data[out_p][t > 0.15, 2:] < 0.001)
    assert_allclose(sim.data[thresholded_p][t > 0.15, 0], 0.6, atol=0.05)
    assert_allclose(sim.data[thresholded_p][t > 0.15, 1], 0.3, atol=0.05)
    assert np.all(sim.data[thresholded_p][t > 0.15, 2:] < 0.001)


def test_thresholding_array_output_shift(Simulator, plt, seed):
    with nengo.Network(seed=seed) as m:
        thresholding = Thresholding(50, 2, 0.5)
        in_node = nengo.Node([0.7, 0.4])
        nengo.Connection(in_node, thresholding.input)
        p = nengo.Probe(thresholding.output, synapse=0.03)

    with Simulator(m) as sim:
        sim.run(0.2)
    t = sim.trange()

    plt.plot(t, sim.data[p])
    plt.axhline(y=0.7)
    plt.ylim(0., 1.)
    plt.xlabel("Time")
    plt.ylabel("Output")

    assert_allclose(sim.data[p][t > 0.15, 0], 0.7, atol=0.1)
    assert_allclose(sim.data[p][:, 1], 0.)


def test_wta(Simulator, plt, seed):
    def input_func(t):
        if t < 0.2:
            return np.array([1.0, 0.8, 0.0, 0.0])
        elif t < 0.3:
            return np.zeros(4)
        else:
            return np.array([0.8, 1.0, 0.0, 0.0])

    with nengo.Network(seed=seed) as m:
        wta = WTA(50, 4, threshold=0.3, function=lambda x: x > 0.)

        in_node = nengo.Node(output=input_func, label='input')
        nengo.Connection(in_node, wta.input)

        in_p = nengo.Probe(in_node)
        out_p = nengo.Probe(wta.output, synapse=0.03)
        thresholded_p = nengo.Probe(wta.thresholded, synapse=0.03)

    with Simulator(m) as sim:
        sim.run(0.5)
    t = sim.trange()
    more_a = (t > 0.15) & (t < 0.2)
    more_b = t > 0.45

    plt.subplot(3, 1, 1)
    plt.plot(t, sim.data[in_p])
    plt.ylabel("Input")
    plt.ylim(0., 1.1)
    plt.subplot(3, 1, 2)
    plt.plot(t, sim.data[out_p])
    plt.plot(t[more_a], np.ones(t.shape)[more_a] * 0.9, c='g', lw=2)
    plt.plot(t[more_b], np.ones(t.shape)[more_b] * 0.9, c='g', lw=2)
    plt.ylabel("Output")
    plt.ylim(0., 1.1)
    plt.subplot(3, 1, 3)
    plt.plot(t, sim.data[thresholded_p])
    plt.plot(t[more_a], np.ones(t.shape)[more_a] * 0.6, c='g', lw=2)
    plt.plot(t[more_b], np.ones(t.shape)[more_b] * 0.6, c='g', lw=2)
    plt.ylabel("Thresholded")
    plt.ylim(0., 1.1)

    assert np.all(sim.data[out_p][more_a, 0] > 0.8)
    assert np.all(sim.data[out_p][more_a, 1] < 0.1)
    assert np.all(sim.data[out_p][more_b, 1] > 0.8)
    assert np.all(sim.data[out_p][more_b, 0] < 0.1)
    assert np.all(sim.data[thresholded_p][more_a, 0] > 0.6)
    assert np.all(sim.data[thresholded_p][more_a, 1:] < 0.001)
    assert np.all(sim.data[thresholded_p][more_b, 1] > 0.6)
    assert np.all(sim.data[thresholded_p][more_b, 0] < 0.001)
    assert np.all(sim.data[thresholded_p][more_b, 2:] < 0.001)



import warnings

import numpy as np

import nengo
from nengo_spa.config import ConfigParam
from nengo.networks.ensemblearray import EnsembleArray
from nengo.params import Default, IntParam, NumberParam
from nengo_spa.network import Network
from nengo.synapses import Lowpass, SynapseParam


# connection weights from (Gurney, Prescott, & Redgrave, 2001)
class Weights(object):
    mm = 1
    mp = 1
    me = 1
    mg = 1
    ws = 1
    wt = 1
    wm = 1
    wg = 1
    wp = 0.9
    we = 0.3
    e = 0.2
    ep = -0.25
    ee = -0.2
    eg = -0.2
    le = 0.2
    lg = 0.2

    @classmethod
    def str_func(cls, x):
        if x < cls.e:
            return 0
        return cls.mm * (x - cls.e)

    @classmethod
    def stn_func(cls, x):
        if x < cls.ep:
            return 0
        return cls.mp * (x - cls.ep)

    @classmethod
    def gpe_func(cls, x):
        if x < cls.ee:
            return 0
        return cls.me * (x - cls.ee)

    @classmethod
    def gpi_func(cls, x):
        if x < cls.eg:
            return 0
        return cls.mg * (x - cls.eg)


class BasalGanglia(Network):
    """Winner take all network, typically used for action selection.

    The basal ganglia network outputs approximately 0 at the dimension with
    the largest value, and is negative elsewhere.

    While the basal ganglia is primarily defined by its winner-take-all
    function, it is also organized to match the organization of the human
    basal ganglia. It consists of five ensembles:

    * Striatal D1 dopamine-receptor neurons (``strD1``)
    * Striatal D2 dopamine-receptor neurons (``strD2``)
    * Subthalamic nucleus (``stn``)
    * Globus pallidus internus / substantia nigra reticulata (``gpi``)
    * Globus pallidus externus (``gpe``)

    Interconnections between these areas are also based on known
    neuroanatomical connections. See [1]_ for more details, and [2]_ for
    the original non-spiking basal ganglia model by
    Gurney, Prescott & Redgrave that this model is based on.

    .. note:: The default `.Solver` for the basal ganglia is `.NnlsL2nz`, which
              requires SciPy. If SciPy is not installed, the global default
              solver will be used instead.

    Parameters
    ----------
    action_count : int
        Number of actions.
    n_neuron_per_ensemble : int, optional (Default: 100)
        Number of neurons in each ensemble in the network.
    output_weight : float, optional (Default: -3.)
        A scaling factor on the output of the basal ganglia
        (specifically on the connection out of the GPi).
    input_bias : float, optional (Default: 0.)
        An amount by which to bias all dimensions of the input node.
        Biasing the input node is important for ensuring that all input
        dimensions are positive and easily comparable.
    general_config : config, optional (Default: None)
        General configuration to be used when specific parameters are not
        overwritten with `ampa_config` or `gaba_config`.
    ampa_config : config, optional (Default: None)
        Configuration for connections corresponding to biological connections
        to AMPA receptors (i.e., connections from STN to to GPi and GPe).
        If None, a default configuration using a 2 ms lowpass synapse
        will be used.
    gaba_config : config, optional (Default: None)
        Configuration for connections corresponding to biological connections
        to GABA receptors (i.e., connections from StrD1 to GPi, StrD2 to GPe,
        and GPe to GPi and STN). If None, a default configuration using an
        8 ms lowpass synapse will be used.
    kwargs
        Passed through the ``spa.Network``.

    Attributes
    ----------
    bias_input : Node or None
        If ``input_bias`` is non-zero, this node will be created to bias
        all of the dimensions of the input signal.
    gpe : EnsembleArray
        Globus pallidus externus ensembles.
    gpi : EnsembleArray
        Globus pallidus internus ensembles.
    input : Node
        Accepts the input signal.
    output : Node
        Provides the output signal.
    stn : EnsembleArray
        Subthalamic nucleus ensembles.
    strD1 : EnsembleArray
        Striatal D1 ensembles.
    strD2 : EnsembleArray
        Striatal D2 ensembles.

    References
    ----------
    .. [1] Stewart, T. C., Choo, X., & Eliasmith, C. (2010).
       Dynamic behaviour of a spiking model of action selection in the
       basal ganglia. In Proceedings of the 10th international conference on
       cognitive modeling (pp. 235-40).
    .. [2] Gurney, K., Prescott, T., & Redgrave, P. (2001).
       A computational model of action selection in the basal
       ganglia. Biological Cybernetics 84, 401-423.
    """

    input_synapse = SynapseParam('input_synapse', default=Lowpass(0.002))
    n_neurons_per_ensemble = IntParam(
        'n_neurons_per_ensemble', default=100, low=1, readonly=True)
    output_weight = NumberParam('output_weight', default=-3., readonly=True)
    input_bias = NumberParam('input_bias', default=0., readonly=True)
    general_config = ConfigParam(
        'general_config', default=None, optional=True, readonly=True)
    ampa_config = ConfigParam(
        'ampa_config', default=None, optional=True, readonly=True)
    gaba_config = ConfigParam(
        'gaba_config', default=None, optional=True, readonly=True)

    def __init__(
            self, action_count, n_neurons_per_ensemble=Default,
            output_weight=Default, input_bias=Default, general_config=Default,
            ampa_config=Default, gaba_config=Default, **kwargs):
        super(BasalGanglia, self).__init__(**kwargs)

        self.action_count = action_count
        self.n_neurons_per_ensemble = n_neurons_per_ensemble
        self.output_weight = output_weight
        self.input_bias = input_bias
        self.general_config = general_config
        self.ampa_config = ampa_config
        self.gaba_config = gaba_config

        self.input_connections = {}
        # Affects all ensembles / connections in the BG
        # unless overwritten with general_config
        config = nengo.Config(nengo.Ensemble, nengo.Connection)
        config[nengo.Ensemble].radius = 1.5
        config[nengo.Ensemble].encoders = nengo.dists.Choice([[1]])
        try:
            # Best, if we have SciPy
            config[nengo.Connection].solver = nengo.solvers.NnlsL2nz()
        except ImportError:
            # Warn if we can't use the better decoder solver.
            if (nengo.Connection not in self.general_config.params or
                    'solver' not in self.general_config[nengo.Connection]):
                warnings.warn("SciPy is not installed, so BasalGanglia will "
                              "use the default decoder solver. Installing "
                              "SciPy may improve BasalGanglia performance.")

        gaba_default = nengo.Config(nengo.Connection)
        gaba_default[nengo.Connection].synapse = nengo.Lowpass(0.008)
        ampa_default = nengo.Config(nengo.Connection)
        ampa_default[nengo.Connection].synapse = nengo.Lowpass(0.002)

        ea_params = {'n_neurons': self.n_neurons_per_ensemble,
                     'n_ensembles': self.action_count}

        with self, config, self.general_config:
            self.strD1 = EnsembleArray(
                label="Striatal D1 neurons",
                intercepts=nengo.dists.Uniform(Weights.e, 1), **ea_params)
            self.strD2 = EnsembleArray(
                label="Striatal D2 neurons",
                intercepts=nengo.dists.Uniform(Weights.e, 1), **ea_params)
            self.stn = EnsembleArray(
                label="Subthalamic nucleus",
                intercepts=nengo.dists.Uniform(Weights.ep, 1), **ea_params)
            self.gpi = EnsembleArray(
                label="Globus pallidus internus",
                intercepts=nengo.dists.Uniform(Weights.eg, 1), **ea_params)
            self.gpe = EnsembleArray(
                label="Globus pallidus externus",
                intercepts=nengo.dists.Uniform(Weights.ee, 1), **ea_params)

            self.input = nengo.Node(label="input", size_in=self.action_count)
            self.output = nengo.Node(label="output", size_in=self.action_count)

            # add bias input (BG performs best in the range 0.5--1.5)
            if abs(self.input_bias) > 0.0:
                self.bias_input = nengo.Node(
                    np.ones(self.action_count) * self.input_bias,
                    label="basal ganglia bias")
                nengo.Connection(self.bias_input, self.input)

            # spread the input to StrD1, StrD2, and STN
            nengo.Connection(self.input, self.strD1.input, synapse=None,
                             transform=Weights.ws * (1 + Weights.lg))
            nengo.Connection(self.input, self.strD2.input, synapse=None,
                             transform=Weights.ws * (1 - Weights.le))
            nengo.Connection(self.input, self.stn.input, synapse=None,
                             transform=Weights.wt)

            # connect the striatum to the GPi and GPe (inhibitory)
            strD1_output = self.strD1.add_output('func_str', Weights.str_func)
            strD2_output = self.strD2.add_output('func_str', Weights.str_func)
            with gaba_default, self.gaba_config:
                nengo.Connection(strD1_output, self.gpi.input,
                                 transform=-Weights.wm)
                nengo.Connection(strD2_output, self.gpe.input,
                                 transform=-Weights.wm)

            # connect the STN to GPi and GPe (broad and excitatory)
            tr = Weights.wp * np.ones((self.action_count, self.action_count))
            stn_output = self.stn.add_output('func_stn', Weights.stn_func)
            with ampa_default, self.ampa_config:
                nengo.Connection(stn_output, self.gpi.input, transform=tr)
                nengo.Connection(stn_output, self.gpe.input, transform=tr)

            # connect the GPe to GPi and STN (inhibitory)
            gpe_output = self.gpe.add_output('func_gpe', Weights.gpe_func)
            with gaba_default, self.gaba_config:
                nengo.Connection(
                    gpe_output, self.gpi.input, transform=-Weights.we)
                nengo.Connection(
                    gpe_output, self.stn.input, transform=-Weights.wg)

            # connect GPi to output (inhibitory)
            gpi_output = self.gpi.add_output('func_gpi', Weights.gpi_func)
            nengo.Connection(gpi_output, self.output, synapse=None,
                             transform=self.output_weight)

    def connect_input(self, source, transform, index):
        self.input_connections[index] = nengo.Connection(
            source, self.input[index], transform=transform,
            synapse=self.input_synapse)



import numpy as np

import nengo
from nengo.params import Default, IntParam
from nengo_spa.network import Network
from nengo_spa.vocab import VocabularyOrDimParam


class Compare(Network):
    """A SPA network for computing the dot product of two inputs.

    Parameters
    ----------
    vocab : Vocabulary or int
        The vocabulary to use to interpret the vector. If an integer is given,
        the default vocabulary of that dimensionality will be used.
    neurons_per_dimension : int, optional (Default: 200)
        Number of neurons to use in each product computation.
    kwargs
        Keyword arguments passed through to ``spa.Network``.
    """

    vocab = VocabularyOrDimParam('vocab', default=None, readonly=True)
    neurons_per_dimension = IntParam(
        'neurons_per_dimension', default=200, low=1, readonly=True)

    def __init__(self, vocab=Default, neurons_per_dimension=Default, **kwargs):
        super(Compare, self).__init__(**kwargs)

        self.vocab = vocab
        self.neurons_per_dimension = neurons_per_dimension

        with self:
            with nengo.Config(nengo.Ensemble) as cfg:
                cfg[nengo.Ensemble].eval_points = nengo.dists.CosineSimilarity(
                    self.vocab.dimensions + 2)
                cfg[nengo.Ensemble].intercepts = nengo.dists.CosineSimilarity(
                    self.vocab.dimensions + 2)
                self.product = nengo.networks.Product(
                    self.neurons_per_dimension, self.vocab.dimensions)
            self.output = nengo.Node(size_in=1, label='output')
            nengo.Connection(self.product.output, self.output,
                             transform=np.ones((1, self.vocab.dimensions)))

        self.input_a = self.product.input_a
        self.input_b = self.product.input_b

        self.inputs = dict(
            input_a=(self.input_a, vocab), input_b=(self.input_b, vocab))
        self.outputs = dict(default=(self.output, None))



class SpaException(Exception):
    """A exception within the SPA subsystem."""


class SpaNetworkError(SpaException, ValueError):
    """An error in how SPA keeps track of networks."""


class SpaParseError(SpaException, ValueError):
    """An error encountered while parsing a SPA expression."""


class SpaTypeError(SpaException, ValueError):
    """The evaluation of types in an SPA expression was invalid."""



import pytest

import nengo
import nengo_spa as spa
from nengo_spa.exceptions import SpaNetworkError

import numpy as np


def test_thalamus_basic(Simulator, plt, seed):
    with nengo.Network(seed=seed) as net:
        bg = spa.BasalGanglia(action_count=5)
        input = nengo.Node([0.8, 0.4, 0.4, 0.4, 0.4], label="input")
        nengo.Connection(input, bg.input, synapse=None)

        thal = spa.Thalamus(action_count=5)
        nengo.Connection(bg.output, thal.input)

        p = nengo.Probe(thal.output, synapse=0.01)

    with Simulator(net) as sim:
        sim.run(0.2)

    t = sim.trange()
    output = np.mean(sim.data[p][t > 0.1], axis=0)

    plt.plot(t, sim.data[p])
    plt.ylabel("Output")

    assert output[0] > 0.8
    assert np.all(output[1:] < 0.01)


@pytest.mark.slow
def test_thalamus(Simulator, plt, seed):
    model = spa.Network(seed=seed)

    with model:
        model.vision = spa.State(vocab=16, neurons_per_dimension=80)
        model.vision2 = spa.State(vocab=16, neurons_per_dimension=80)
        model.motor = spa.State(vocab=16, neurons_per_dimension=80)
        model.motor2 = spa.State(vocab=32, neurons_per_dimension=80)

        spa.Actions(
            'dot(vision, A) --> motor=A, motor2=translate(vision*vision2)',
            'dot(vision, B) --> motor=vision, motor2=translate(vision*A*~B)',
            'dot(vision, ~A) --> motor=~vision, '
            'motor2=translate(~vision*vision2)'
        ).build()

        def input_f(t):
            if t < 0.1:
                return 'A'
            elif t < 0.3:
                return 'B'
            elif t < 0.5:
                return '~A'
            else:
                return '0'
        model.input = spa.Input()
        model.input.vision = input_f
        model.input.vision2 = 'B*~A'

        input, vocab = model.get_network_input('motor')
        input2, vocab2 = model.get_network_input('motor2')
        p = nengo.Probe(input, 'output', synapse=0.03)
        p2 = nengo.Probe(input2, 'output', synapse=0.03)

    with Simulator(model) as sim:
        sim.run(0.5)

    t = sim.trange()
    data = vocab.dot(sim.data[p].T)
    data2 = vocab2.dot(sim.data[p2].T)

    plt.subplot(2, 1, 1)
    plt.plot(t, data.T)
    plt.subplot(2, 1, 2)
    plt.plot(t, data2.T)

    # Action 1
    assert data[0, t == 0.1] > 0.8
    assert data[1, t == 0.1] < 0.2
    assert data2[0, t == 0.1] < 0.35
    assert data2[1, t == 0.1] > 0.4
    # Action 2
    assert data[0, t == 0.3] < 0.2
    assert data[1, t == 0.3] > 0.8
    assert data2[0, t == 0.3] > 0.5
    assert data2[1, t == 0.3] < 0.3
    # Action 3
    assert data[0, t == 0.5] > 0.8
    assert data[1, t == 0.5] < 0.2
    assert data2[0, t == 0.5] < 0.5
    assert data2[1, t == 0.5] > 0.4


def test_routing(Simulator, seed, plt):
    model = spa.Network(seed=seed)
    model.config[spa.State].vocab = 3
    model.config[spa.State].subdimensions = 3
    with model:
        model.ctrl = spa.State(16, subdimensions=16, label='ctrl')

        def input_func(t):
            if t < 0.2:
                return 'A'
            elif t < 0.4:
                return 'B'
            else:
                return 'C'
        model.input = spa.Input()
        model.input.ctrl = input_func

        model.buff1 = spa.State(label='buff1')
        model.buff2 = spa.State(label='buff2')
        model.buff3 = spa.State(label='buff3')

        node1 = nengo.Node([0, 1, 0])
        node2 = nengo.Node([0, 0, 1])

        nengo.Connection(node1, model.buff1.input)
        nengo.Connection(node2, model.buff2.input)

        spa.Actions(
            'dot(ctrl, A) --> buff3=buff1',
            'dot(ctrl, B) --> buff3=buff2',
            'dot(ctrl, C) --> buff3=buff1*buff2',
        ).build()

        buff3_probe = nengo.Probe(model.buff3.output, synapse=0.03)

    with Simulator(model) as sim:
        sim.run(0.6)

    data = sim.data[buff3_probe]

    plt.plot(sim.trange(), data)

    valueA = np.mean(data[150:200], axis=0)  # should be [0, 1, 0]
    valueB = np.mean(data[350:400], axis=0)  # should be [0, 0, 1]
    valueC = np.mean(data[550:600], axis=0)  # should be [1, 0, 0]

    assert valueA[0] < 0.2
    assert valueA[1] > 0.75
    assert valueA[2] < 0.2

    assert valueB[0] < 0.2
    assert valueB[1] < 0.2
    assert valueB[2] > 0.75

    assert valueC[0] > 0.75
    assert valueC[1] < 0.2
    assert valueC[2] < 0.2


def test_routing_recurrency_compilation(Simulator, seed, plt):
    model = spa.Network(seed=seed)
    model.config[spa.State].vocab = 2
    model.config[spa.State].subdimensions = 2
    with model:
        model.buff1 = spa.State(label='buff1')
        model.buff2 = spa.State(label='buff2')
        spa.Actions('0.5 --> buff2=buff1, buff1=buff2').build()

    with Simulator(model) as sim:
        assert sim


def test_nondefault_routing(Simulator, seed):
    model = spa.Network(seed=seed)
    model.config[spa.State].vocab = 3
    model.config[spa.State].subdimensions = 3
    with model:
        model.ctrl = spa.State(16, subdimensions=16, label='ctrl')

        def input_func(t):
            if t < 0.2:
                return 'A'
            elif t < 0.4:
                return 'B'
            else:
                return 'C'
        model.input = spa.Input()
        model.input.ctrl = input_func

        model.buff1 = spa.State(label='buff1')
        model.buff2 = spa.State(label='buff2')
        model.cmp = spa.Compare(3)

        node1 = nengo.Node([0, 1, 0])
        node2 = nengo.Node([0, 0, 1])

        nengo.Connection(node1, model.buff1.input)
        nengo.Connection(node2, model.buff2.input)

        spa.Actions(
            'dot(ctrl, A) --> cmp.input_a=buff1, cmp.input_b=buff1',
            'dot(ctrl, B) --> cmp.input_a=buff1, cmp.input_b=buff2',
            'dot(ctrl, C) --> cmp.input_a=buff2, cmp.input_b=buff2',
        ).build()

        compare_probe = nengo.Probe(model.cmp.output, synapse=0.03)

    with Simulator(model) as sim:
        sim.run(0.6)

    similarity = sim.data[compare_probe]

    valueA = np.mean(similarity[150:200], axis=0)  # should be [1]
    valueB = np.mean(similarity[350:400], axis=0)  # should be [0]
    valueC = np.mean(similarity[550:600], axis=0)  # should be [1]

    assert valueA > 0.6
    assert valueB < 0.3
    assert valueC > 0.6


def test_errors():
    # motor does not exist
    with pytest.raises(SpaNetworkError):
        with spa.Network() as model:
            model.vision = spa.State(vocab=16)
            spa.Actions('0.5 --> motor=A').build()


def test_constructed_objects_are_accessible():
    with spa.Network() as model:
        model.config[spa.State].vocab = 16
        model.state1 = spa.State()
        model.state2 = spa.State()
        model.state3 = spa.State()

        actions = spa.Actions(
            'dot(state1, A) --> state2 = state3',
            '0.5 --> state2 = B')
        bg, thalamus, _ = actions.build()

        print(thalamus.fixed_connections)
        print(thalamus.gate_out_connections)
        assert isinstance(thalamus.gates[0], nengo.Ensemble)
        assert isinstance(thalamus.gate_in_connections[0], nengo.Connection)
        assert isinstance(thalamus.gate_out_connections[0], nengo.Connection)
        assert isinstance(thalamus.channels[0], spa.State)
        assert isinstance(
            thalamus.channel_out_connections[0], nengo.Connection)

        assert isinstance(thalamus.fixed_connections[1], nengo.Connection)

        assert thalamus.bg_connection.pre is bg.output
        assert thalamus.bg_connection.post is thalamus.input



import nengo
from nengo.exceptions import ValidationError
from nengo.networks.ensemblearray import EnsembleArray
from nengo.params import BoolParam, Default, IntParam, NumberParam
from nengo_spa.network import Network
from nengo_spa.vocab import VocabularyOrDimParam
from nengo_spa.networks import IdentityEnsembleArray


class State(Network):
    """A SPA network capable of representing a single vector, with optional
    memory.

    This is a minimal SPA network, useful for passing data along (for example,
    visual input).

    Parameters
    ----------
    vocab : Vocabulary or int
        The vocabulary to use to interpret the vector. If an integer is given,
        the default vocabulary of that dimensionality will be used.
    subdimensions : int, optional (Default: 16)
        Size of the individual ensembles making up the vector.
        Must divide ``dimensions`` evenly.
    neurons_per_dimensions : int, optional (Default: 50)
        Number of neurons in an ensemble will be
        ``neurons_per_dimensions * subdimensions``.
    feedback : float, optional (Default: 0.0)
        Gain of feedback connection. Set to 1.0 for perfect memory,
        or 0.0 for no memory. Values in between will create a decaying memory.
    feedback_synapse : float, optional (Default: 0.1)
        The synapse on the feedback connection.
    kwargs
        Keyword arguments passed through to ``spa.Network``.
    """

    vocab = VocabularyOrDimParam('vocab', default=None, readonly=True)
    subdimensions = IntParam('subdimensions', default=16, low=1, readonly=True)
    neurons_per_dimension = IntParam(
        'neurons_per_dimension', default=50, low=1, readonly=True)
    feedback = NumberParam('feedback', default=.0, readonly=True)
    feedback_synapse = NumberParam(
        'feedback_synapse', default=.1, readonly=True)
    represent_identity = BoolParam(
        'represent_identity', default=True, readonly=True)

    def __init__(self, vocab=Default, subdimensions=Default,
                 neurons_per_dimension=Default, feedback=Default,
                 represent_identity=Default,
                 feedback_synapse=Default, **kwargs):
        super(State, self).__init__(**kwargs)

        self.vocab = vocab
        self.subdimensions = subdimensions
        self.neurons_per_dimension = neurons_per_dimension
        self.feedback = feedback
        self.feedback_synapse = feedback_synapse
        self.represent_identity = represent_identity

        dimensions = self.vocab.dimensions

        if dimensions % self.subdimensions != 0:
            raise ValidationError(
                "Dimensions (%d) must be divisible by subdimensions (%d)" % (
                    dimensions, self.subdimensions),
                attr='dimensions', obj=self)

        with self:
            if self.represent_identity:
                self.state_ensembles = IdentityEnsembleArray(
                    self.neurons_per_dimension, dimensions, self.subdimensions,
                    label='state')
            else:
                self.state_ensembles = EnsembleArray(
                    self.neurons_per_dimension * self.subdimensions,
                    dimensions // self.subdimensions,
                    ens_dimensions=self.subdimensions,
                    eval_points=nengo.dists.CosineSimilarity(dimensions + 2),
                    intercepts=nengo.dists.CosineSimilarity(dimensions + 2),
                    label='state')

            if self.feedback is not None and self.feedback != 0.0:
                nengo.Connection(
                    self.state_ensembles.output, self.state_ensembles.input,
                    transform=self.feedback, synapse=self.feedback_synapse)

        self.input = self.state_ensembles.input
        self.output = self.state_ensembles.output
        self.inputs = dict(default=(self.input, self.vocab))
        self.outputs = dict(default=(self.output, self.vocab))



"""Abstract syntax trees for SPA actions.

SPA actions are parsed into an abstract syntax tree (AST) representing the
structure of the action expressions. In the first step, this syntax tree is
independent of the actual model; it only represents the syntactic information.

For the construction of the AST the Python eval function is used (by modifiying
the name lookup in the globals dictionary). Because of this all Python code
not involving identifiers will be statically evaluated before insertion into
the syntax tree (e.g., '2 * 3 + 1' will be inserted as ``Scalar(7)``).

Each node in the syntax tree will evaluate to a specific type. The most
important types are ``TScalar`` for expressions evaluating to a scalar and
:class:`TVocabulary` for expressions evaluating to a semantic pointer. This
latter type describes the vocabulary the semantic pointer belongs to and
different vocabularies give different types. This ensures that only semantic
pointers of a matching vocabulary are assigned.

To determine the type of each node an actual model has to be provided. This is
because names of semantic pointers are not associated with a vocabulary and it
needs to be inferred from some actual SPA network for which we have to be able
to resolve the names of those networks. There are a few basic rules for this
type inference:

1. If something with unknown vocabulary is assigned to a network, that
   network's vocabulary provides the type.
2. If a binary operation has an operand with unknown vocabulary it is
   determined from the other operand.

Once all the types have been determined, the AST can be used to construct
Nengo objects to perform the operations represented with the AST. In this
process each node in the syntax tree can create :class:`Artifact`s. These
give generated Nengo objects to be connected to the appropriate places
including the transform that should be used on the connection. This is
necessary because at the time most objects are constructed we only know this
constructed object and the transform, but not what it is supposed to connect
to. So the final connection will be done by some other node in the syntax tree.

To avoid confusion with the normal Nengo build process, we use the term
'construct' here.


In the following we provide the grammar of SPA actions for reference:

``
Scalar: <any Python expression evaluating to a single number != 0>
Symbol: <valid Python identifier starting with a capital letter>
Zero: '0'
Module: <valid Python identifier> | <valid Python identifier> '.' Module
Source: S'(' Source ')' | Scalar | Symbol | Zero | Module | BinaryOperation |
        UnaryOperation | DotProduct
BinaryOperation: Product | Sum | Difference
Product: Source '*' Source
Sum: Source '+' Source
Difference: Source '-' Source
UnaryOperation: ApproxInverse | Negative
ApproxInverse: '~' Source
Negative: '-' Source
DotProduct: 'dot(' Source ',' Source ')'
Reinterpret: 'reinterpret(' Source (',' VocabArg)? ')'
Translate: 'translate(' Source (',' VocabArg)? ')'
VocabArg: 'vocab='? (Module | <valid Python identifier>)
Sink: <valid Python identifier> | <valid Python identifier> '.' Sink
Effect: Sink '=' Source
Effects: Effect | Effect ',' Effects
Action: Source '-->' Effects
``

Note that `Difference` ``a - b`` will be represented as `a + (-b)` in the AST.

Operator precedence is defined as follows from highest to lowest priority:

``
0 Scalar, Symbol, Zero, Module, DotProduct, Reinterpret, Translate
1 UnaryOperation
2 Product
3 Sum
``
"""
import warnings
from collections import defaultdict

import numpy as np

import nengo
from nengo_spa.modules.bind import Bind
from nengo_spa.modules.compare import Compare
from nengo_spa.pointer import SemanticPointer
from nengo_spa.modules.product import Product as ProductModule
from nengo_spa.exceptions import SpaNetworkError, SpaParseError, SpaTypeError


class ConstructionContext(object):
    """Context in which SPA actions are constructed.

    This primarily provides the SPA networks used to construct certain
    components. All attributes except `root_network` may be ``None`` if these
    are not provided in the current construction context.

    Attributes
    ----------
    root_network : :class:`spa.Module`
        The root network the encapsulated all of the constructed structures.
    bg : :class:`spa.BasalGanglia`
        Module to manage the basal ganglia part of action selection.
    thalamus : :class:`spa.Thalamus`
        Module to manage the thalamus part of action selection.
    sink : :class:`Sink`
        Node in the AST where some result will be send to.
    active_net : class:`nengo.Network`
        Network to add constructed components to.
    """
    __slots__ = [
        'root_network', 'bg', 'thalamus', 'bias', 'sink', 'active_net',
        'constructed']

    def __init__(
            self, root_network, bg=None, thalamus=None,
            sink=None, active_net=None, constructed=None):
        self.root_network = root_network
        self.bg = bg
        self.thalamus = thalamus
        self.bias = None
        self.sink = sink
        if active_net is None:
            active_net = root_network
        self.active_net = active_net
        if constructed is None:
            constructed = defaultdict(list)
        self.constructed = constructed

    def subcontext(self, sink=None, active_net=None):
        if sink is None:
            sink = self.sink
        if active_net is None:
            active_net = self.active_net
        return self.__class__(
            root_network=self.root_network, bg=self.bg,
            thalamus=self.thalamus, sink=sink, active_net=active_net,
            constructed=self.constructed)

    @property
    def sink_network(self):
        return self.root_network.get_spa_network(
            self.sink.name, strip_output=True)

    @property
    def sink_input(self):
        return self.root_network.get_network_input(self.sink.name)

    def add_constructed(self, ast_node, *objects):
        self.constructed[ast_node].extend(objects)


class Type(object):
    """Describes a type.

    Each part of the AST evaluates to some type.
    """
    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return '{}({!r})'.format(self.__class__.__name__, self.name)

    def __str__(self):
        return self.name

    def __hash__(self):
        return hash(self.__class__) ^ hash(self.name)

    def __eq__(self, other):
        return self.__class__ is other.__class__ and self.name == other.name

    def __ne__(self, other):
        return not self == other


TAction = Type('TAction')
TScalar = Type('TScalar')
TEffect = Type('TEffect')
TEffects = Type('TEffects')


class TVocabulary(Type):
    """Each vocabulary is treated as its own type.

    All vocabulary types constitute a type class.
    """
    def __init__(self, vocab):
        super(TVocabulary, self).__init__('TVocabulary')
        self.vocab = vocab

    def __repr__(self):
        return '{}({!r}, {!r})'.format(
            self.__class__.__name__, self.name, self.vocab)

    def __str__(self):
        return '{}<{}>'.format(self.name, self.vocab)

    def __hash__(self):
        return super(TVocabulary, self).__hash__() ^ hash(self.vocab)

    def __eq__(self, other):
        return (super(TVocabulary, self).__eq__(other) and
                self.vocab is other.vocab)


class Artifact(object):
    """Stores information about Nengo objects constructed from SPA actions.

    This deals with the problem that when we construct the Nengo object we
    have the object itself and know the transform for the outgoing connection,
    but we do not know what to connect to yet. Thus, this class allows to store
    and pass around that information until we know what to connect to.

    Attributes
    ----------
    nengo_source : :class:`nengo.NengoObject`
        Some constructed Nengo object that allows outgoing connections.
    transform : array-like
        Transform to be applied to the outgoing connection from the
        `nengo_source`.
    """
    def __init__(self, nengo_source, transform=1):
        self.nengo_source = nengo_source
        self.transform = transform

    def add_transform(self, tr):
        return Artifact(self.nengo_source, np.dot(tr, self.transform))


def infer_vocab(root_network, *nodes):
    """Return the first vocabulary type that can be inferred for one of the
    `nodes`.

    If the context that an operation is embedded in does not provide a
    vocabulary type, it might be possible to infer it from one of the nodes
    in the operation.

    If no vocabulary type can be inferred, ``None`` will be returned.

    Note that this function calls ``infer_types`` on a subset or all `nodes`
    which has side effects!
    """
    for node in nodes:
        try:
            node.infer_types(root_network, None)
            if isinstance(node.type, TVocabulary):
                return node.type
        except SpaTypeError:
            pass
    return None


def construct_bias(ast_node, value, context):
    """Constructs a bias node (if not existent) and a transform to `value`."""
    with context.active_net:
        if context.bias is None:
            context.bias = nengo.Node([1], label="bias")
            context.add_constructed(ast_node, context.bias)
    if isinstance(value, SemanticPointer):
        value = value.v
    transform = np.array([value]).T
    return [Artifact(context.bias, transform=transform)]


def value_to_transform(value):
    if isinstance(value, SemanticPointer):
        value = np.atleast_2d(value.v).T
    return np.asarray(value)


def ensure_node(obj):
    """Return `obj` as :class:`Node`.

    Leaves `obj` unchanged if it is already an instance of :class:`Node`.
    """
    if not isinstance(obj, Node):
        if obj is 0:
            obj = Zero()
        else:
            obj = Scalar(obj)
    return obj


class Node(object):
    """Abstract class for a node in the AST.

    Attributes
    ----------
    staticity : int
        Staticity of the node. See `.Node.Staticity` which also lists valid
        values.
    type : :class:`Type`
        Type that this node evaluates to. This will be set to ``None`` until
        the type inference was run.
    """

    class Staticity:
        """Valid staticity values.

        * ``FIXED``: Value of the node is static, i.e. does not change over
          time.
        * ``TRANSFORM_ONLY``: Value of the node changes over time, but can be
          implemented with a transform on existing neural resources.
        * ``DYNAMIC``: Value of the node is fully dynamic and needs additional
          neural resources to be implemented.
        """
        FIXED = 0
        TRANSFORM_ONLY = 1
        DYNAMIC = 2

    def __init__(self, staticity, precedence=0):
        self.staticity = staticity
        self.precedence = precedence
        self.type = None

    @property
    def fixed(self):
        """Indicates whether the node value is static.

        A static node value does not change over time.
        """
        return self.staticity <= self.Staticity.FIXED

    def __hash__(self):
        h = hash(self.__class__)
        for v in self.__dict__.values():
            h ^= hash(v)
        return h

    def __eq__(self, other):
        return (self.__class__ is other.__class__ and
                self.__dict__ == other.__dict__)

    def infer_types(self, root_network, context_type):
        """Run type inference on this node and its children.

        Will raise a :class:`nengo.exceptions.SpaTypeError` if invalid,
        non-matching, or undefined types are discovered.

        This function is idempotent.

        Parameters
        ----------
        root_network : :class:`spa.Module`
            The root_network used to resolve names.
        context_type : :class:`Type`
            The type of the context of this node. Allows to infer the type
            from the context if the node has no definitive type on its own.
        """
        raise NotImplementedError()

    def construct(self, context):
        """Construct network components for this node and its children.

        Parameters
        ----------
        context : :class:`ConstructionContext`
            The context in which the network components are constructed.

        Returns
        -------
        list of :class:`Articfact`
            The constructed objects with transforms that should be connected to
            the objects that continue the processing.
        """
        raise NotImplementedError()

    def evaluate(self):
        """Evaluate the value of this node statically.

        This can only be done for ``fixed`` nodes, otherwise a ``ValueError``
        will be raised.
        """
        raise NotImplementedError()


class Source(Node):
    """Abstract base class of all AST nodes that can provide some output value.
    """
    def __invert__(self):
        return ApproxInverse(self)

    def __neg__(self):
        return Negative(self)

    def __add__(self, other):
        return Sum(self, other)

    def __radd__(self, other):
        return Sum(other, self)

    def __sub__(self, other):
        return Sum(self, Negative(other))

    def __rsub__(self, other):
        return Sum(other, Negative(self))

    def __mul__(self, other):
        return Product(self, other)

    def __rmul__(self, other):
        return Product(other, self)

    def infer_types(self, root_network, context_type):
        raise NotImplementedError()

    def construct(self, context):
        raise NotImplementedError()

    def evaluate(self):
        raise NotImplementedError()


class Scalar(Source):
    """A fixed scalar."""
    def __init__(self, value):
        super(Scalar, self).__init__(staticity=Node.Staticity.FIXED)
        self.value = value
        self.type = TScalar

    def infer_types(self, root_network, context_type):
        pass

    def construct(self, context):
        return construct_bias(self, self.value, context)

    def evaluate(self):
        return self.value

    def __str__(self):
        return str(self.value)


class Symbol(Source):
    """A fixed semantic pointer identified by its name (`key`).

    The `key` has to start with a capatial letter.
    """
    def __init__(self, key):
        super(Symbol, self).__init__(staticity=Node.Staticity.FIXED)
        self.validate(key)
        self.key = key

    def validate(self, key):
        if not key[0].isupper():
            raise SpaParseError(
                "Semantic pointers must begin with a capital letter.")

    def infer_types(self, root_network, context_type):
        if not isinstance(context_type, TVocabulary):
            raise SpaTypeError("Invalid type.")
        self.type = context_type
        # Make sure that key has been added to vocab after type inference to
        # make translate operations as deterministic as possible (it could
        # differ if at a later point another key would be added).
        self.type.vocab.parse(self.key)

    def construct(self, context):
        value = self.type.vocab[self.key].v
        return construct_bias(self, value, context)

    def evaluate(self):
        return self.type.vocab[self.key]

    def __str__(self):
        return self.key


class Zero(Source):
    """Zero which can act as scalar or zero vector."""

    def __init__(self):
        super(Zero, self).__init__(staticity=Node.Staticity.FIXED)

    def infer_types(self, root_network, context_type):
        if context_type is None:
            self.type = TScalar
        elif context_type == TScalar or isinstance(context_type, TVocabulary):
            self.type = context_type
        else:
            raise SpaTypeError("Invalid type.")

    def construct(self, context):
        return []

    def evaluate(self):
        if self.type == TScalar:
            return 0
        else:
            return SemanticPointer(np.zeros(self.type.vocab.dimensions))

    def __str__(self):
        return "0"


class Module(Source):
    """A SPA network or network output identified by its name.

    This will provide potentially time varying input. This class is not used
    for networks that act as sink.
    """
    def __init__(self, name):
        super(Module, self).__init__(staticity=Node.Staticity.TRANSFORM_ONLY)
        self.name = name

    def infer_types(self, root_network, context_type):
        vocab = root_network.get_network_output(self.name)[1]
        if vocab is None:
            self.type = TScalar
        else:
            self.type = TVocabulary(vocab)

    def construct(self, context):
        return [Artifact(
            context.root_network.get_network_output(self.name)[0])]

    def evaluate(self):
        raise ValueError("Module cannot be statically evaluated.")

    def __getattr__(self, name):
        return Module(self.name + '.' + name)

    def __str__(self):
        return self.name


class BinaryNode(Source):
    """Base class for binary operators.

    Attributes
    ----------
    lhs : :class:`Node`
        Left-hand side
    rhs : :class:`Node`
        Right-hand side
    """
    def __init__(self, lhs, rhs, staticity, precedence=0):
        lhs = ensure_node(lhs)
        rhs = ensure_node(rhs)

        super(BinaryNode, self).__init__(
            staticity=staticity, precedence=precedence)
        self.lhs = lhs
        self.rhs = rhs

    def infer_types(self, root_network, context_type):
        raise NotImplementedError()

    def construct(self, context):
        raise NotImplementedError()

    def _connect_binary_operation(self, context, net):
        with context.root_network:
            for artifact in self.lhs.construct(context):
                context.add_constructed(
                    self, nengo.Connection(
                        artifact.nengo_source, net.input_a,
                        transform=artifact.transform))
            for artifact in self.rhs.construct(context):
                context.add_constructed(
                    self, nengo.Connection(
                        artifact.nengo_source, net.input_b,
                        transform=artifact.transform))


class BinaryOperation(BinaryNode):
    """Base class for binary operators.

    Attributes
    ----------
    lhs : :class:`Node`
        Left-hand side
    rhs : :class:`Node`
        Right-hand side
    operator : str
        String representation of the operator.
    """
    def __init__(
            self, lhs, rhs, operator, staticity, precedence=2,
            allow_scalar=False):
        super(BinaryOperation, self).__init__(
            lhs, rhs, staticity, precedence=precedence)
        self.operator = operator
        self.allow_scalar = allow_scalar

    def infer_types(self, root_network, context_type):
        if context_type is None:
            context_type = infer_vocab(root_network, self.lhs, self.rhs)

        self.lhs.infer_types(root_network, context_type)
        self.rhs.infer_types(root_network, context_type)

        if self.lhs.type == self.rhs.type:
            self.type = self.lhs.type
        elif self.allow_scalar and self.lhs.type == TScalar:
            self.type = self.rhs.type
        elif self.allow_scalar and self.rhs.type == TScalar:
            self.type = self.lhs.type
        else:
            raise SpaTypeError(
                "Incompatible types {} and {} in operation '{}'.".format(
                    self.lhs.type, self.rhs.type, self))

    def construct(self, context):
        raise NotImplementedError()

    def __str__(self):
        if self.lhs.precedence > self.precedence:
            lhs_str = '({})'.format(self.lhs)
        else:
            lhs_str = str(self.lhs)
        if self.rhs.precedence > self.precedence:
            rhs_str = '({})'.format(self.rhs)
        else:
            rhs_str = str(self.rhs)

        return '{} {} {}'.format(lhs_str, self.operator, rhs_str)


class Product(BinaryOperation):
    def __init__(self, lhs, rhs):
        lhs = ensure_node(lhs)
        rhs = ensure_node(rhs)

        if not lhs.fixed and not rhs.fixed:
            staticity = Node.Staticity.DYNAMIC
        else:
            staticity = max(lhs.staticity, rhs.staticity)

        super(Product, self).__init__(
            lhs, rhs, '*', staticity, allow_scalar=True)

    def construct(self, context):
        if self.fixed:
            return construct_bias(self, self.evaluate(), context)

        if self.lhs.fixed:
            tr = self.lhs.evaluate()
            artifacts = self.rhs.construct(context)
        if self.rhs.fixed:
            tr = self.rhs.evaluate()
            artifacts = self.lhs.construct(context)

        is_binding = (isinstance(self.lhs.type, TVocabulary) and
                      isinstance(self.rhs.type, TVocabulary))

        if self.lhs.fixed or self.rhs.fixed:
            if is_binding:
                tr = tr.get_convolution_matrix()
            else:
                tr = value_to_transform(tr)
            return [x.add_transform(tr) for x in artifacts]

        with context.active_net:
            if is_binding:
                net = Bind(self.type.vocab, label=str(self))
            elif self.lhs.type == TScalar and self.rhs.type == TScalar:
                net = ProductModule()
            else:
                raise NotImplementedError(
                    "Dynamic scaling of semantic pointer not implemented.")
        context.add_constructed(self, net)

        self._connect_binary_operation(context, net)
        return [Artifact(net.output)]

    def evaluate(self):
        return self.lhs.evaluate() * self.rhs.evaluate()


class Sum(BinaryOperation):
    def __init__(self, lhs, rhs):
        lhs = ensure_node(lhs)
        rhs = ensure_node(rhs)
        staticity = min(
            Node.Staticity.TRANSFORM_ONLY, max(lhs.staticity, rhs.staticity))
        super(Sum, self).__init__(lhs, rhs, '+', staticity, precedence=3)

    def construct(self, context):
        if self.fixed:
            return construct_bias(self, self.evaluate(), context)

        return (self.lhs.construct(context) +
                self.rhs.construct(context))

    def evaluate(self):
        return self.lhs.evaluate() + self.rhs.evaluate()


class UnaryOperation(Source):
    """Base class for unary operators.

    Attributes
    ----------
    source : :class:`Node`
        Node the operator is applied to.
    operator : str
        String representation of the operator.
    """
    def __init__(self, source, operator, precedence=1):
        source = ensure_node(source)
        super(UnaryOperation, self).__init__(
            staticity=source.staticity, precedence=precedence)
        self.source = source
        self.operator = operator

    def infer_types(self, root_network, context_type):
        self.source.infer_types(root_network, context_type)
        self.type = self.source.type

    def construct(self, context):
        raise NotImplementedError()

    def __str__(self):
        if self.source.precedence <= self.precedence:
            return self.operator + str(self.source)
        else:
            return self.operator + '(' + str(self.source) + ')'


class ApproxInverse(UnaryOperation):
    def __init__(self, source):
        super(ApproxInverse, self).__init__(source, '~')

    def infer_types(self, root_network, context_type):
        super(ApproxInverse, self).infer_types(root_network, context_type)
        if not isinstance(self.type, TVocabulary):
            raise SpaTypeError(
                "Cannot apply approximate inverse to '{}' which is not of "
                "type TVocabulary, but {}.".format(self.source, self.type))

    def construct(self, context):
        if self.fixed:
            return construct_bias(self, self.evaluate(), context)

        d = self.type.vocab.dimensions
        tr = np.eye(d)[-np.arange(d)]
        return [x.add_transform(tr) for x in self.source.construct(context)]

    def evaluate(self):
        return ~self.source.evaluate()


class Negative(UnaryOperation):
    def __init__(self, source):
        super(Negative, self).__init__(source, '-')

    def construct(self, context):
        if self.fixed:
            return construct_bias(self, self.evaluate(), context)
        return [x.add_transform(-1) for x in self.source.construct(context)]

    def evaluate(self):
        return -self.source.evaluate()


class DotProduct(BinaryNode):
    def __init__(self, lhs, rhs):
        lhs = ensure_node(lhs)
        rhs = ensure_node(rhs)

        if not lhs.fixed and not rhs.fixed:
            staticity = Node.Staticity.DYNAMIC
        else:
            staticity = max(lhs.staticity, rhs.staticity)

        super(DotProduct, self).__init__(lhs, rhs, staticity)
        self.type = TScalar

    def infer_types(self, root_network, context_type):
        context_type = infer_vocab(root_network, self.lhs, self.rhs)
        self.lhs.infer_types(root_network, context_type)
        self.rhs.infer_types(root_network, context_type)
        if not isinstance(self.lhs.type, TVocabulary):
            raise SpaTypeError(
                "First argument of dot product '{}' is not of type "
                "TVocabulary, but {}.".format(self, self.lhs.type))
        if not isinstance(self.rhs.type, TVocabulary):
            raise SpaTypeError(
                "Second argument of dot product '{}' is not of type "
                "TVocabulary, but {}.".format(self, self.rhs.type))
        if self.lhs.type.vocab is not self.rhs.type.vocab:
            raise SpaTypeError(
                "Incompatible types {} and {} in dot product '{}'.".format(
                    self.lhs.type, self.rhs.type, self))

    def construct(self, context):
        if self.fixed:
            return construct_bias(self, self.evaluate(), context)

        if self.lhs.fixed:
            tr = value_to_transform(self.lhs.evaluate()).T
            return [x.add_transform(tr)
                    for x in self.rhs.construct(context)]
        if self.rhs.fixed:
            tr = value_to_transform(self.rhs.evaluate()).T
            return [x.add_transform(tr)
                    for x in self.lhs.construct(context)]

        assert self.lhs.type.vocab is self.rhs.type.vocab
        with context.active_net:
            net = Compare(self.lhs.type.vocab, label=str(self))
            self._connect_binary_operation(context, net)
        context.add_constructed(self, net)
        return [Artifact(net.output)]

    def evaluate(self):
        return np.dot(self.lhs.evaluate(), self.rhs.evaluate())

    def __str__(self):
        return 'dot({}, {})'.format(self.lhs, self.rhs)


class Reinterpret(Source):
    def __init__(self, source, vocab=None):
        source = ensure_node(source)
        super(Reinterpret, self).__init__(staticity=source.staticity)
        self.source = source
        self.vocab = vocab

    def infer_types(self, root_network, context_type):
        if self.vocab is None:
            self.type = context_type
        elif isinstance(self.vocab, Module):
            self.vocab.infer_types(root_network, None)
            self.type = self.vocab.type
        else:
            self.type = TVocabulary(self.vocab)
        if not isinstance(self.type, TVocabulary):
            raise SpaTypeError(
                "Cannot infer vocabulary for '{}'.".format(self))

        self.source.infer_types(root_network, None)
        if not isinstance(self.source.type, TVocabulary):
            raise SpaTypeError(
                "Cannot reinterpret '{}' because it is not of type "
                "TVocabulary, but {}.".format(self.source, self.source.type))
        if self.source.type.vocab.dimensions != self.type.vocab.dimensions:
            raise SpaTypeError(
                "Cannot reinterpret '{}' with {}-dimensional vocabulary as "
                "{}-dimensional vocabulary.".format(
                    self.source, self.source.type.vocab.dimensions,
                    self.type.vocab.dimensions))

    def construct(self, context):
        return self.source.construct(context)

    def evaluate(self):
        return self.source.evaluate()

    def __str__(self):
        return 'reinterpret({})'.format(self.source)


class Translate(Source):
    def __init__(self, source, vocab=None, populate=None):
        source = ensure_node(source)
        super(Translate, self).__init__(staticity=source.staticity)
        self.source = source
        self.vocab = vocab
        self.populate = populate

    def infer_types(self, root_network, context_type):
        if self.vocab is None:
            self.type = context_type
        elif isinstance(self.vocab, Module):
            self.vocab.infer_types(root_network, None)
            self.type = self.vocab.type
        else:
            self.type = TVocabulary(self.vocab)
        if not isinstance(self.type, TVocabulary):
            raise SpaTypeError(
                "Cannot infer vocabulary for '{}'.".format(self))

        self.source.infer_types(root_network, None)
        if not isinstance(self.source.type, TVocabulary):
            raise SpaTypeError(
                "Cannot translate '{}' because it is not of type "
                "TVocabulary, but {}.".format(self.source, self.source.type))

    def construct(self, context):
        tr = self.source.type.vocab.transform_to(
            self.type.vocab, populate=self.populate)
        artifacts = self.source.construct(context)
        return [a.add_transform(tr) for a in artifacts]

    def evaluate(self):
        tr = self.source.type.vocab.transform_to(
            self.type.vocab, populate=self.populate)
        return SemanticPointer(np.dot(tr, self.source.evaluate().v))

    def __str__(self):
        return 'translate({})'.format(self.source)


class Effect(Node):
    """Assignment of an expression to a SPA network.

    Attributes
    ----------
    sink : :class:`Sink`
        Module that acts as sink.
    source : :class:`Source`
        Source of information to be fed to the sink.
    channeled : bool
        Indicates if information should be passed through an additional
        (inhibitable) channel between the source and sink.
    channel : :class:`nengo.networks.EnsembleArray`
        The channel that was constructed for this effect. Will initially be
        ``None`` and will only be constructed if `channeled` is ``True``.
    """
    def __init__(self, sink, source, channeled=False):
        source = ensure_node(source)
        super(Effect, self).__init__(staticity=source.staticity)
        self.type = TEffect
        self.sink = ensure_node(sink)
        self.source = ensure_node(source)
        self.channeled = channeled
        self.channel = None

    def infer_types(self, root_network, context_type):
        if context_type is not None:
            raise ValueError("Effect only allows a context_type of None.")
        self.sink.infer_types(root_network, None)
        self.source.infer_types(root_network, self.sink.type)
        if self.sink.type != self.source.type:
            raise SpaTypeError("Cannot assign {} to {} in '{}'".format(
                self.source.type, self.sink.type, self))

    def construct(self, context):
        assert context.sink is None
        context = context.subcontext(sink=self.sink)

        if self.channeled and self.fixed:
            return []  # Will be implemented in transform from thalamus

        if self.channeled:
            self.channel = context.thalamus.construct_channel(
                context.sink_network, context.sink_input,
                net=context.active_net, label='channel: ' + str(self))
            target = self.channel.input
            connect_fn = context.thalamus.connect
        else:
            target = context.sink_input[0]
            connect_fn = nengo.Connection

        artifacts = self.source.construct(context)
        for artifact in artifacts:
            context.add_constructed(self, connect_fn(
                artifact.nengo_source, target, transform=artifact.transform))
        return []

    def evaluate(self):
        raise ValueError("Effects cannot be statically evaluated.")

    def __str__(self):
        return '{} = {}'.format(self.sink, self.source)


class Effects(Node):
    """Multiple effects."""
    def __init__(self, *effects):
        super(Effects, self).__init__(
            staticity=max(e.staticity for e in effects))
        self.type = TEffects
        self.effects = effects

    def infer_types(self, root_network, context_type):
        for e in self.effects:
            e.infer_types(root_network, context_type)

    def construct(self, context):
        for effect in self.effects:
            effect.construct(context)
        return []

    def evaluate(self):
        raise ValueError("Effects cannot be statically evaluated.")

    def __str__(self):
        return ', '.join(str(e) for e in self.effects)


class Sink(Node):
    """SPA network that acts as sink identified by its name."""

    def __init__(self, name):
        super(Sink, self).__init__(staticity=Node.Staticity.DYNAMIC)
        self.name = name

    def infer_types(self, root_network, context_type):
        vocab = root_network.get_network_input(self.name)[1]
        if vocab is None:
            self.type = TScalar
        else:
            self.type = TVocabulary(vocab)

    def construct(self, context):
        return []

    def evaluate(self):
        raise ValueError("Sinks cannot be statically evaluated.")

    def __str__(self):
        return self.name


class Action(Node):
    """A conditional SPA action.

    Attributes
    ----------
    index : int
        Numerical index of the action.
    condition : :class:`Node`
        Condition for the action's effects to become active.
    effets : :class:`Node`
        Effects when the condition is met.
    name : str
        Name of the action.
    """
    def __init__(self, condition, effects, index=0, name=None):
        super(Action, self).__init__(staticity=Node.Staticity.DYNAMIC)
        self.type = TAction
        self.index = index
        self.condition = ensure_node(condition)
        self.effects = ensure_node(effects)
        self.name = name

    @property
    def effect(self):
        warnings.warn(DeprecationWarning("Use the effects attribute instead."))
        return self.effects

    def infer_types(self, root_network, context_type):
        if isinstance(self.condition, Node):
            self.condition.infer_types(root_network, context_type)
            if self.condition.type != TScalar:
                raise SpaTypeError(
                    "Condition '{}' does not evaluate to a scalar.".format(
                        self.condition))

        self.effects.infer_types(root_network, None)

    def construct(self, context):
        if context.bg is None or context.thalamus is None:
            raise SpaNetworkError(
                "Conditional actions require basal ganglia and thalamus.")

        # construct bg utility
        if self.condition.staticity <= Node.Staticity.TRANSFORM_ONLY:
            condition_context = context
        else:
            condition_context = context.subcontext(active_net=nengo.Network(
                label='condition: ' + str(self.condition)))
        condition_artifacts = self.condition.construct(condition_context)

        # construct effects
        self.effects.construct(context)

        # construct thalamus gate
        if not self.effects.fixed:
            gate_label = 'gate[{}]: {}'.format(self.index, self.condition)
            context.thalamus.construct_gate(
                self.index, net=context.thalamus, label=gate_label)

        for artifact in condition_artifacts:
            context.bg.connect_input(
                artifact.nengo_source, artifact.transform, self.index)

        # connect up
        for effect in self.effects.effects:
            if effect.fixed:
                sink = context.root_network.get_network_input(
                    effect.sink.name)[0]
                tr = value_to_transform(effect.source.evaluate())
                context.thalamus.connect_fixed(self.index, sink, transform=tr)
            else:
                context.thalamus.connect_gate(self.index, effect.channel)

        return []

    def evaluate(self):
        raise NotImplementedError("Cannot evaluate conditional actions.")

    def __str__(self):
        return '{} --> {}'.format(self.condition, self.effects)



# -*- coding: utf-8 -*-

import numpy as np
import pytest

from nengo.exceptions import NengoWarning, SpaParseError, ValidationError
from nengo_spa import Vocabulary, VocabularyMap
from nengo_spa.pointer import Identity
from nengo.utils.testing import warns


def test_add(rng):
    v = Vocabulary(3, rng=rng)
    v.add('A', [1, 2, 3])
    v.add('B', [4, 5, 6])
    v.add('C', [7, 8, 9])
    assert np.allclose(v.vectors, [[1, 2, 3], [4, 5, 6], [7, 8, 9]])


def test_populate(rng):
    v = Vocabulary(64, rng=rng)

    v.populate('A')
    assert 'A' in v

    v.populate('B; C')
    assert 'B' in v
    assert 'C' in v

    v.populate('D.unitary()')
    assert 'D' in v
    np.testing.assert_almost_equal(np.linalg.norm(v['D'].v), 1.)
    np.testing.assert_almost_equal(np.linalg.norm((v['D'] * v['D']).v), 1.)

    v.populate('E = A + 2 * B')
    assert np.allclose(v['E'].v, v.parse('A + 2 * B').v)
    assert np.linalg.norm(v['E'].v) > 2.

    v.populate('F = (A + 2 * B).normalized()')
    assert np.allclose(v['F'].v, v.parse('A + 2 * B').normalized().v)
    np.testing.assert_almost_equal(np.linalg.norm(v['F'].v), 1.)

    v.populate('G = A; H')
    assert np.allclose(v['G'].v, v['A'].v)
    assert 'H' in v

    # Assigning non-existing pointer
    with pytest.raises(NameError):
        v.populate('I = J')

    # Redefining
    with pytest.raises(ValidationError):
        v.populate('H = A')

    # Calling non existing function
    with pytest.raises(AttributeError):
        v.populate('I = H.invalid()')

    # invalid names: lowercase, unicode
    with pytest.raises(SpaParseError):
        v.populate('x = A')
    # with pytest.raises(SpaParseError):
    v.populate(u'Aα = A')


def test_parse(rng):
    v = Vocabulary(64, rng=rng)
    v.populate('A; B; C')
    A = v.parse('A')
    B = v.parse('B')
    C = v.parse('C')
    assert np.allclose((A * B).v, v.parse('A * B').v)
    assert np.allclose((A * ~B).v, v.parse('A * ~B').v)
    assert np.allclose((A + B).v, v.parse('A + B').v)
    assert np.allclose((A - (B*C)*3 + ~C).v, v.parse('A-(B*C)*3+~C').v)

    assert np.allclose(v.parse('0').v, np.zeros(64))
    assert np.allclose(v.parse('1').v, np.eye(64)[0])
    assert np.allclose(v.parse('1.7').v, np.eye(64)[0] * 1.7)

    with pytest.raises(SyntaxError):
        v.parse('A((')
    with pytest.raises(SpaParseError):
        v.parse('"hello"')
    with pytest.raises(SpaParseError):
        v.parse('"hello"')


def test_parse_n(rng):
    v = Vocabulary(64, rng=rng)
    v.populate('A; B; C')
    A = v.parse('A')
    B = v.parse('B')

    parsed = v.parse_n('A', 'A*B', 'A+B', '3')
    assert np.allclose(parsed[0].v, A.v)
    assert np.allclose(parsed[1].v, (A * B).v)
    assert np.allclose(parsed[2].v, (A + B).v)
    assert np.allclose(parsed[3].v, 3 * Identity(64).v)


def test_invalid_dimensions():
    with pytest.raises(ValidationError):
        Vocabulary(1.5)
    with pytest.raises(ValidationError):
        Vocabulary(0)
    with pytest.raises(ValidationError):
        Vocabulary(-1)


def test_capital(rng):
    v = Vocabulary(16, rng=rng)
    with pytest.raises(SpaParseError):
        v.parse('a')
    with pytest.raises(SpaParseError):
        v.parse('A+B+C+a')


def test_transform(recwarn, rng):
    v1 = Vocabulary(32, strict=False, rng=rng)
    v2 = Vocabulary(64, strict=False, rng=rng)
    v1.populate('A; B; C')
    v2.populate('A; B; C')
    A = v1['A']
    B = v1['B']
    C = v1['C']

    # Test transform from v1 to v2 (full vocbulary)
    # Expected: np.dot(t, A.v) ~= v2.parse('A')
    # Expected: np.dot(t, B.v) ~= v2.parse('B')
    # Expected: np.dot(t, C.v) ~= v2.parse('C')
    t = v1.transform_to(v2)

    assert v2.parse('A').compare(np.dot(t, A.v)) > 0.9
    assert v2.parse('C+B').compare(np.dot(t, C.v + B.v)) > 0.9

    # Test transform from v1 to v2 (only 'A' and 'B')
    t = v1.transform_to(v2, keys=['A', 'B'])

    assert v2.parse('A').compare(np.dot(t, A.v)) > 0.9
    assert v2.parse('B').compare(np.dot(t, C.v + B.v)) > 0.9

    # Test warns on missing keys
    v1.populate('D')
    D = v1['D']
    with warns(NengoWarning):
        v1.transform_to(v2)

    # Test populating missing keys
    t = v1.transform_to(v2, populate=True)
    assert v2.parse('D').compare(np.dot(t, D.v)) > 0.9

    # Test ignores missing keys in source vocab
    v2.populate('E')
    v1.transform_to(v2, populate=True)
    assert 'E' not in v1


def test_create_pointer_warning(rng):
    v = Vocabulary(2, rng=rng)

    # five pointers shouldn't fit
    with warns(UserWarning):
        v.populate('A; B; C; D; E')


def test_readonly(rng):
    v1 = Vocabulary(32, rng=rng)
    v1.populate('A;B;C')

    v1.readonly = True

    with pytest.raises(ValueError):
        v1.parse('D')


def test_subset(rng):
    v1 = Vocabulary(32, rng=rng)
    v1.populate('A; B; C; D; E; F; G')

    # Test creating a vocabulary subset
    v2 = v1.create_subset(['A', 'C', 'E'])
    assert list(v2.keys()) == ['A', 'C', 'E']
    assert v2['A'] == v1['A']
    assert v2['C'] == v1['C']
    assert v2['E'] == v1['E']


def test_vocabulary_set(rng):
    v8 = Vocabulary(8)
    v16 = Vocabulary(16)
    v32 = Vocabulary(32)
    vs = VocabularyMap([v8, v16], rng=rng)

    # Behaviour common to set and dict
    assert len(vs) == 2
    assert 8 in vs
    assert 16 in vs
    assert 32 not in vs

    assert v8 in vs
    assert v16 in vs
    assert v32 not in vs
    assert Vocabulary(8) not in vs

    # dict behaviour
    assert vs[8] is v8
    assert vs[16] is v16

    del vs[8]
    assert 8 not in vs

    # set behaviour
    vs.add(v32)
    assert vs[32] is v32
    with pytest.warns(UserWarning):
        vs.add(v32)

    vs.discard(32)
    assert 32 not in vs
    vs.discard(v16)
    assert 16 not in vs

    # creating new vocabs if non existent
    vs.add(v8)
    assert vs.get_or_create(8) is v8
    new = vs.get_or_create(16)
    assert vs[16] is new
    assert new.dimensions == 16
    assert new.rng is rng



import re

import numpy as np
from numpy.testing import assert_almost_equal

from nengo_spa.examine import pairs, similarity, text
from nengo_spa.vocab import Vocabulary


def test_similarity(rng):
    v = Vocabulary(64, max_similarity=0.1)
    v.populate('''
        A;
        B;
        C = .7 * A + .3 * B;
        D = C.normalized()
    ''')

    s = similarity(v['A'].v, v)
    assert_almost_equal(s[0], 1.)
    assert s[1] < 0.1
    assert_almost_equal(s[2], 0.7 + .3 * np.dot(v['A'].v, v['B'].v))

    sn = similarity(v['A'].v, v, normalize=True)
    assert_almost_equal(s[3], sn[2])

    data = np.array([v['A'].v, v['B'].v])
    s = similarity(data, v)
    assert_almost_equal(s[0, 0], 1.)
    assert_almost_equal(s[1, 1], 1.)
    assert s[0, 1] < 0.1
    assert s[1, 0] < 0.1

    s = similarity(data, v.parse_n('A', 'B'))
    assert_almost_equal(s[0, 0], 1.)
    assert_almost_equal(s[1, 1], 1.)
    assert s[0, 1] < 0.1
    assert s[1, 0] < 0.1


def test_pairs():
    v = Vocabulary(64)
    v.populate('A; B; C')
    actual = pairs(v)
    expected = {'A*B', 'A*C', 'B*C'}
    assert actual == expected


def test_text(rng):
    v = Vocabulary(64, rng=rng)
    v.populate('A; B; C; D; E; F')
    x = v.parse('A+B+C')
    y = v.parse('-D-E-F')
    ptr = r'-?[01]\.[0-9]{2}[A-F]'
    assert re.match(';'.join([ptr] * 3), text(x, v))
    assert re.match(';'.join([ptr] * 2), text(x, v, maximum_count=2))
    assert re.match(ptr, text(x, v, maximum_count=1))
    assert len(text(x, v, maximum_count=10).split(';')) <= 10
    assert re.match(';'.join([ptr] * 4), text(x, v, minimum_count=4))
    assert re.match(';'.join([ptr.replace('F', 'C')] * 3),
                    text(x, v, minimum_count=4, terms=['A', 'B', 'C']))

    assert re.match(ptr, text(y, v, threshold=0.6))
    assert text(y, v, minimum_count=None, threshold=0.6) == ''

    assert text(x, v, join=',') == text(x, v).replace(';', ',')
    assert re.match(';'.join([ptr] * 2), text(x, v, normalize=True))

    assert text([0]*64, v) == '0.00F'
    assert text(v['D'].v, v) == '1.00D'



"""Selection networks that pick one or more options among multiple choices."""
import nengo
import numpy as np


def IA(
        n_neurons, n_ensembles, accum_threshold=0.8, accum_neuron_ratio=0.7,
        accum_timescale=0.1, feedback_timescale=0.005,
        accum_synapse=0.1, ff_synapse=0.005,
        intercept_width=0.15, radius=1., **kwargs):
    """Independent accumulator (IA) winner-take-all (WTA) network.

    This is a two-layered network. The first layer consists of independent
    accumulators (integrators), whereas the second layer does a thresholding.
    Once the threshold is exceeded a feedback connection will stabilize the
    current choice and inhibit all other choices. To switch the selection,
    it is necessary to provide a transient input to *input_reset* to reset
    the accumulator states.

    This network is suited especially for accumulating evidence under noisy
    conditions and keep a stable choice selection until the processing of the
    choice has been finished.

    Further details are to be found in "A Spiking Independent Accumulator Model
    for Winner-Take-All Computation" (J Gosmann, AR Voelker, and C Eliasimth;
    submitted to CogSci 2017).

    Parameters
    ----------
    n_neurons : int
        Number of neurons for each choice.
    n_ensembles : int
        Number of choices.
    accum_threshold : float, optional (Default: 0.8)
        Accumulation threshold that needs to be reached to produce an output.
    accum_neuron_ratio: float, optional (Default: 0.7)
        Portion of *n_neurons* that will be used for a layer 1 accumulator
        ensemble. The remaining neurons will be used for a layer 2 thresholding
        ensemble.
    accum_timescale : float, optional (Default: 0.1)
        Evidence accumulation timescale.
    feedback_timescale : float, optional (Default: 0.005)
        Timescale for the feedback connection from the thresholding layer to
        the accumulation layer.
    accum_synapse : Synapse or float, optional (Default: 0.1)
        The synapse for connections to the accumulator ensembles.
    ff_synapse : Synapse or float, optional (Default: 0.005)
        Synapse for feed-forward connections.
    intercept_width : float, optional (Default: 0.15)
        The `presets.ThresholdingEnsembles` *intercept_width* parameter.
    radius : float, optional (Default: 1.0)
        The representational radius of the ensembles.
    kwargs : dict
        Passed on to `nengo.Network`.

    Attributes
    ----------
    input : Node
        The inputs to the network.
    input_reset : Node
        Input to reset the accumulators.
    output : Node
        The outputs of the network.
    accumulators : Thresholding
        The layer 1 accumulators.
    thresholding : Thresholding
        The layer 2 thresholding ensembles.
    """
    n_accum_neurons = int(accum_neuron_ratio * n_neurons)
    n_thresholding_neurons = n_neurons - n_accum_neurons

    bar_beta = 1. + radius * feedback_timescale / accum_timescale
    feedback_tr = (
        np.eye(n_ensembles) - bar_beta * (1. - np.eye(n_ensembles)) /
        feedback_timescale)

    with nengo.Network(**kwargs) as net:
        net.accumulators = Thresholding(
            n_accum_neurons, n_ensembles, threshold=0., radius=radius)
        net.thresholding = Thresholding(
            n_thresholding_neurons, n_ensembles, threshold=accum_threshold,
            radius=radius, function=lambda x: x > accum_threshold)

        nengo.Connection(
            net.accumulators.output, net.accumulators.input,
            synapse=accum_synapse)
        nengo.Connection(
            net.accumulators.output, net.thresholding.input,
            synapse=ff_synapse)
        nengo.Connection(
            net.thresholding.output, net.accumulators.input,
            synapse=accum_synapse, transform=accum_synapse * feedback_tr)

        net.input_reset = nengo.Node(size_in=1)
        nengo.Connection(
            net.input_reset, net.accumulators.input, synapse=None,
            transform=-radius * np.ones((n_ensembles, 1)) / accum_synapse)

        net.input = nengo.Node(size_in=n_ensembles)
        nengo.Connection(net.input, net.accumulators.input, synapse=None,
                         transform=1. / accum_timescale)
        net.output = net.thresholding.output
    return net


def Thresholding(
        n_neurons, n_ensembles, threshold, intercept_width=0.15, function=None,
        radius=1., **kwargs):
    """Array of thresholding ensembles.

    All inputs below the threshold will produce an output of 0, whereas inputs
    above the threshold produce an output of equal value.

    Parameters
    ----------
    n_neurons : int
        Number of neurons for each ensemble.
    n_ensembles : int
        Number of ensembles.
    threshold : float
        The thresholding value.
    intercept_width : float, optional (Default: 0.15)
        The `presets.ThresholdingEnsembles` *intercept_width* parameter.
    function : function, optional (Default: None)
        Function to apply to the thresholded values.
    radius : float, optional (Default: 1.0)
        The representational radius of the ensembles.
    kwargs : dict
        Arguments passed on to `nengo.Network`.

    Attributes
    ----------
    input : Node
        The inputs to the network.
    output : Node
        The outputs of the network.
    thresholded : Node
        The raw thresholded value (before applying *function* or correcting
        for the shift produced by the thresholding).
    """
    with nengo.Network(**kwargs) as net:
        with nengo.presets.ThresholdingEnsembles(
                0., intercept_width, radius=radius):
            net.thresholding = nengo.networks.EnsembleArray(
                n_neurons, n_ensembles)

        net.bias = nengo.Node(1.)
        nengo.Connection(net.bias, net.thresholding.input,
                         transform=-threshold * np.ones((n_ensembles, 1)))

        net.input = net.thresholding.input
        net.thresholded = net.thresholding.output

        if function is None:
            function = lambda x: x
        function = lambda x, function=function: function(x + threshold)
        net.output = net.thresholding.add_output('function', function)
    return net


def WTA(n_neurons, n_ensembles, inhibit_scale=1.0, inhibit_synapse=0.005,
        **kwargs):
    """Winner-take-all (WTA) network with lateral inhibition.

    Parameters
    ----------
    n_neurons : int
        Number of neurons for each ensemble.
    n_ensembles : int
        Number of ensembles.
    inhibit_scale : float, optional (Default: 1.0)
        Scaling of the lateral inhibition.
    inhibit_synapse : Synapse or float, optional (Default: 0.005)
        Synapse on the recurrent connection for lateral inhibition.
    kwargs : dict
        Arguments passed on to `Thresholding`.

    Attributes
    ----------
    input : Node
        The inputs to the network.
    output : Node
        The outputs of the network.
    thresholded : Node
        The raw thresholded value (before applying *function* or correcting
        for the shift produced by the thresholding).
    """
    net = Thresholding(n_neurons, n_ensembles, **kwargs)
    with net:
        nengo.Connection(
            net.thresholded, net.input,
            transform=inhibit_scale * (np.eye(n_ensembles) - 1.),
            synapse=inhibit_synapse)
    return net



import warnings

import nengo
from nengo.exceptions import ValidationError
from nengo.utils.network import with_self


class IdentityEnsembleArray(nengo.Network):
    """An ensemble array optimized for representing the identity vector for
    circular convolution.

    The ensemble array will use ensembles with *subdimensions* dimensions,
    except for the first *subdimensions* dimensions. These will be split into
    a one-dimensional ensemble for the first dimension and a *subdimensions-1*
    dimensional ensemble.

    Parameters
    ----------
    neurons_per_dimension : int
        Neurons per dimension.
    dimensions : int
        Total number of dimensions. Must be a multiple of *subdimensions*.
    subdimensions : int
        Maximum number of dimensions per ensemble.
    kwargs : dict
        Arguments to pass through to the `nengo.Network` constructor.

    Attributes
    ----------
    input : nengo.Node
        Input node.
    output : nengo.Node
        Output node.
    """
    def __init__(
            self, neurons_per_dimension, dimensions, subdimensions, **kwargs):
        super(IdentityEnsembleArray, self).__init__(**kwargs)

        self.neurons_per_dimension = neurons_per_dimension
        self.dimensions = dimensions
        self.subdimensions = subdimensions

        self.neuron_input = None

        cos_sim_dist = nengo.dists.CosineSimilarity(dimensions + 2)
        with self:
            self.input = nengo.Node(size_in=dimensions)
            self.output = nengo.Node(size_in=dimensions)

            first = nengo.Ensemble(neurons_per_dimension, 1)
            nengo.Connection(self.input[0], first, synapse=None)
            nengo.Connection(first, self.output[0], synapse=None)

            if subdimensions > 1:
                second = nengo.Ensemble(
                    neurons_per_dimension * (subdimensions - 1),
                    subdimensions - 1,
                    eval_points=cos_sim_dist, intercepts=cos_sim_dist)
                nengo.Connection(
                    self.input[1:subdimensions], second, synapse=None)
                nengo.Connection(
                    second, self.output[1:subdimensions], synapse=None)

            if dimensions > subdimensions:
                remainder = nengo.networks.EnsembleArray(
                    neurons_per_dimension * subdimensions,
                    dimensions // subdimensions - 1, subdimensions,
                    eval_points=cos_sim_dist, intercepts=cos_sim_dist)
                nengo.Connection(
                    self.input[subdimensions:], remainder.input, synapse=None)
                nengo.Connection(
                    remainder.output, self.output[subdimensions:],
                    synapse=None)

    @with_self
    def add_neuron_input(self):
        """Adds a node that provides input to the neurons of all ensembles.

        This node is accessible through the *neuron_input* attribute.

        Returns
        -------
        nengo.Node
            The added node.
        """
        if self.neuron_input is not None:
            warnings.warn("neuron_input already exists. Returning.")
            return self.neuron_input

        if any(isinstance(e.neuron_type, nengo.Direct)
               for e in self.all_ensembles):
            raise ValidationError(
                "Ensembles use Direct neuron type. "
                "Cannot give neuron input to Direct neurons.",
                attr='all_ensembles[i].neuron_type', obj=self)

        self.neuron_input = nengo.Node(
            size_in=self.neurons_per_dimension * self.dimensions,
            label="neuron_input")

        i = 0
        for ens in self.all_ensembles:
            nengo.Connection(
                self.neuron_input[i:(i + ens.n_neurons)], ens.neurons,
                synapse=None)
            i += ens.n_neurons

        return self.neuron_input



import numpy as np

import nengo
import nengo_spa as spa


def test_basic(Simulator, plt, seed):
    with nengo.Network(seed=seed) as model:
        bg = spa.BasalGanglia(action_count=5, seed=seed)
        input = nengo.Node([0.8, 0.4, 0.4, 0.4, 0.4], label="input")
        nengo.Connection(input, bg.input, synapse=None)
        p = nengo.Probe(bg.output, synapse=0.01)

    with Simulator(model) as sim:
        sim.run(0.2)

    t = sim.trange()
    output = np.mean(sim.data[p][t > 0.1], axis=0)

    plt.plot(t, sim.data[p])
    plt.ylabel("Output")

    assert output[0] > -0.1
    assert np.all(output[1:] < -0.8)


def test_basal_ganglia(Simulator, seed, plt):
    model = spa.Network(seed=seed)

    with model:
        model.vision = spa.State(vocab=16)
        model.motor = spa.State(vocab=16)
        model.compare = spa.Compare(vocab=16)

        # test all acceptable condition formats
        spa.Actions(
            '0.5 --> motor=A',
            'dot(vision, CAT) --> motor=B',
            'dot(vision*CAT, DOG) --> motor=C',
            '2*dot(vision, CAT*0.5) --> motor=D',
            'dot(vision, CAT) + 0.5 - dot(vision,CAT) --> motor=E',
            'dot(vision, PARROT) + compare --> motor=F',
            '0.5*dot(vision, MOUSE) + 0.5*compare --> motor=G',
            '( dot(vision, MOUSE) - compare ) * 0.5 --> motor=H'
        ).build()

        def input(t):
            if t < 0.1:
                return '0'
            elif t < 0.2:
                return 'CAT'
            elif t < 0.3:
                return 'DOG*~CAT'
            elif t < 0.4:
                return 'PARROT'
            elif t < 0.5:
                return 'MOUSE'
            else:
                return '0'
        model.input = spa.Input()
        model.input.vision = input
        model.input.compare.input_a = 'SHOOP'
        model.input.compare.input_b = 'SHOOP'
        p = nengo.Probe(model.bg.input, 'output', synapse=0.03)

    with Simulator(model) as sim:
        sim.run(0.5)
    t = sim.trange()

    plt.plot(t, sim.data[p])
    plt.legend(["A", "B", "C", "D", "E", "F", "G", "H"])
    plt.title('Basal Ganglia output')

    # assert the basal ganglia is prioritizing things correctly
    # Motor F
    assert sim.data[p][t == 0.4, 5] > 0.8
    # Motor G
    assert sim.data[p][t == 0.5, 6] > 0.8
    # Motor A
    assert 0.6 > sim.data[p][t == 0.1, 0] > 0.4
    # Motor B
    assert sim.data[p][t == 0.2, 1] > 0.8
    # Motor C
    assert sim.data[p][t == 0.3, 2] > 0.6

    # Motor B should be the same as Motor D
    assert np.allclose(sim.data[p][:, 1], sim.data[p][:, 3])
    # Motor A should be the same as Motor E
    assert np.allclose(sim.data[p][:, 0], sim.data[p][:, 4])


def test_scalar_product():
    with spa.Network() as model:
        model.scalar = spa.Scalar()
        spa.Actions('scalar*scalar --> scalar=1').build()
    # just testing network construction without exception here


def test_constructed_input_connections_are_accessible():
    with spa.Network() as model:
        model.config[spa.State].vocab = 16
        model.state1 = spa.State()
        model.state2 = spa.State()

        actions = spa.Actions('dot(state1, A) --> state2 = A')
        bg, thalamus, _ = actions.build()

        assert isinstance(bg.input_connections[0], nengo.Connection)



import numpy as np
import pytest

from nengo.exceptions import ValidationError
from nengo_spa.pointer import AbsorbingElement, Identity, SemanticPointer, Zero
from nengo.utils.compat import range


def test_init():
    a = SemanticPointer([1, 2, 3, 4])
    assert len(a) == 4

    a = SemanticPointer([1, 2, 3, 4, 5])
    assert len(a) == 5

    a = SemanticPointer(list(range(100)))
    assert len(a) == 100

    a = SemanticPointer(27)
    assert len(a) == 27
    assert np.allclose(a.length(), 1)

    with pytest.raises(ValidationError):
        a = SemanticPointer(np.zeros((2, 2)))

    with pytest.raises(ValidationError):
        a = SemanticPointer(-1)
    with pytest.raises(ValidationError):
        a = SemanticPointer(0)
    with pytest.raises(ValidationError):
        a = SemanticPointer(1.7)
    with pytest.raises(ValidationError):
        a = SemanticPointer(None)
    with pytest.raises(TypeError):
        a = SemanticPointer(int)


def test_length():
    a = SemanticPointer([1, 1])
    assert np.allclose(a.length(), np.sqrt(2))
    a = SemanticPointer(10)*1.2
    assert np.allclose(a.length(), 1.2)


def test_normalized():
    a = SemanticPointer([1, 1]).normalized()
    b = a.normalized()
    assert a is not b
    assert np.allclose(b.length(), 1)


def test_str():
    a = SemanticPointer([1, 1])
    assert str(a) == '[ 1.  1.]'


def test_make_unitary(rng):
    a = SemanticPointer(100, rng=rng)
    b = a.unitary()
    assert a is not b
    assert np.allclose(1, b.length())
    assert np.allclose(1, (b * b).length())
    assert np.allclose(1, (b * b * b).length())


def test_add_sub():
    a = SemanticPointer(10)
    b = SemanticPointer(10)
    c = a.copy()
    d = b.copy()

    c += b
    d -= -a

    assert np.allclose((a + b).v, a.v + b.v)
    assert np.allclose((a + b).v, c.v)
    assert np.allclose((a + b).v, d.v)
    assert np.allclose((a + b).v, (a - (-b)).v)


def test_convolution(rng):
    a = SemanticPointer(64, rng=rng)
    b = SemanticPointer(64, rng=rng)
    identity = SemanticPointer(np.eye(64)[0])

    c = a.copy()
    c *= b

    ans = np.fft.irfft(np.fft.rfft(a.v) * np.fft.rfft(b.v))

    assert np.allclose((a * b).v, ans)
    assert np.allclose(a.convolve(b).v, ans)
    assert np.allclose(c.v, ans)
    assert np.allclose((a * identity).v, a.v)
    assert (a * b * ~b).compare(a) > 0.6


def test_multiply():
    a = SemanticPointer(50)

    assert np.allclose((a * 5).v, a.v * 5)
    assert np.allclose((5 * a).v, a.v * 5)
    assert np.allclose((a * 5.7).v, a.v * 5.7)
    assert np.allclose((5.7 * a).v, a.v * 5.7)
    assert np.allclose((0 * a).v, np.zeros(50))
    assert np.allclose((1 * a).v, a.v)

    with pytest.raises(Exception):
        a * None
    with pytest.raises(Exception):
        a * 'string'


def test_compare(rng):
    a = SemanticPointer(50, rng=rng) * 10
    b = SemanticPointer(50, rng=rng) * 0.1

    assert a.compare(a) > 0.99
    assert a.compare(b) < 0.2
    assert np.allclose(a.compare(b), a.dot(b) / (a.length() * b.length()))


def test_dot(rng):
    a = SemanticPointer(50, rng=rng) * 1.1
    b = SemanticPointer(50, rng=rng) * (-1.5)
    assert np.allclose(a.dot(b), np.dot(a.v, b.v))


def test_distance(rng):
    a = SemanticPointer(50, rng=rng)
    b = SemanticPointer(50, rng=rng)
    assert a.distance(a) < 1e-5
    assert a.distance(b) > 0.85


def test_invert():
    a = SemanticPointer(50)
    assert a.v[0] == (~a).v[0]
    assert a.v[49] == (~a).v[1]
    assert np.allclose(a.v[1:], (~a).v[:0:-1])


def test_len():
    a = SemanticPointer(5)
    assert len(a) == 5

    a = SemanticPointer(list(range(10)))
    assert len(a) == 10


def test_copy():
    a = SemanticPointer(5)
    b = a.copy()
    assert a is not b
    assert a.v is not b.v
    assert np.allclose(a.v, b.v)


def test_mse():
    a = SemanticPointer(50)
    b = SemanticPointer(50)

    assert np.allclose(((a - b).length() ** 2) / 50, a.mse(b))


def test_conv_matrix():
    a = SemanticPointer(50)
    b = SemanticPointer(50)

    m = b.get_convolution_matrix()

    assert np.allclose((a*b).v, np.dot(m, a.v))


def test_identity(rng):
    p = Identity(64)
    assert len(p) == 64
    assert np.sum(p.v) == 1.
    a = SemanticPointer(64, rng)
    assert np.allclose(a.v, (a * p).v)


def test_absorbing_element(rng):
    p = AbsorbingElement(64)
    assert len(p) == 64
    assert p.length() == 1.
    a = SemanticPointer(64, rng)
    assert np.allclose(p.v, np.abs((a * p).normalized().v))
    assert np.allclose(p.v, np.abs((a * -p).normalized().v))


def test_zero():
    z = Zero(64)
    assert len(z) == 64
    assert np.all(z.v == 0.)



"""Basic networks that are used by nengo_spa.

These networks do not provide any information about inputs, outputs or used
vocabularies and are completely independent of SPA specifics.
"""

from .circularconvolution import CircularConvolution
from .identity_ensemble_array import IdentityEnsembleArray



from nengo.config import Config
from nengo.exceptions import ValidationError
from nengo.params import Parameter


class ConfigParam(Parameter):
    """A parameter where the value is a `.Config` object."""

    def validate(self, instance, value):
        super(ConfigParam, self).validate(instance, value)

        if value is not None and not isinstance(value, Config):
            raise ValidationError(
                "Must be of type 'Config' (got type %r)."
                % type(value).__name__, attr=self.name, obj=instance)

    def __set__(self, instance, value):
        if value is None:
            value = Config()
        super(ConfigParam, self).__set__(instance, value)



from collections import Mapping
import re
import warnings

import numpy as np

import nengo
from nengo.exceptions import NengoWarning, SpaParseError, ValidationError
from nengo_spa import pointer
from nengo_spa.pointer import Identity
from nengo.utils.compat import is_number, is_integer, range


valid_sp_regex = re.compile('[A-Z][_a-zA-Z0-9]*')


class Vocabulary(Mapping):
    """A collection of semantic pointers, each with their own text label.

    The Vocabulary can also act as a dictionary, with keys as the names
    of the semantic pointers and values as the `.SemanticPointer` objects
    themselves. If it is asked for a pointer that does not exist, one will
    be automatically created.

    Parameters
    -----------
    dimensions : int
        Number of dimensions for each semantic pointer.
    strict : bool, optional (Default: True)
        TODO
    max_similarity : float, optional (Default: 0.1)
        When randomly generating pointers, ensure that the cosine of the
        angle between the new pointer and all existing pointers is less
        than this amount. If the system is unable to find such a pointer
        after 100 tries, a warning message is printed.
    rng : `numpy.random.RandomState`, optional (Default: None)
        The random number generator to use to create new vectors.

    Attributes
    ----------
    keys : list of strings
        The names of all known semantic pointers (e.g., ``['A', 'B', 'C']``).
    vectors : ndarray
        All of the semantic pointer values in a matrix, in the same order
        as in ``keys``.
    """

    def __init__(self, dimensions, strict=True, max_similarity=0.1, rng=None):

        if not is_integer(dimensions) or dimensions < 1:
            raise ValidationError("dimensions must be a positive integer",
                                  attr='dimensions', obj=self)
        self.dimensions = dimensions
        self.strict = strict
        self.max_similarity = max_similarity
        self.pointers = {}
        self._keys = []
        self._vectors = np.zeros((0, dimensions), dtype=float)
        self.rng = rng

    @property
    def vectors(self):
        v = self._vectors.view()
        v.setflags(write=False)
        return v

    def __str__(self):
        return '{}-dimensional vocab at 0x{:x}'.format(
            self.dimensions, id(self))

    def create_pointer(self, attempts=100, transform=None):
        """Create a new semantic pointer.

        This will take into account the max_similarity
        parameter from self. If a pointer satisfying max_similarity
        is not generated after the specified number of attempts, the
        candidate pointer with lowest maximum cosine with all existing
        pointers is returned.
        """
        if len(self) == 0:
            best_p = pointer.SemanticPointer(self.dimensions, rng=self.rng)
        else:
            best_p = None
            best_sim = np.inf
            for _ in range(attempts):
                p = pointer.SemanticPointer(self.dimensions, rng=self.rng)
                if transform is not None:
                    p = eval('p.' + transform, {'p': p}, self)
                p_sim = np.max(np.dot(self._vectors, p.v))
                if p_sim < best_sim:
                    best_p = p
                    best_sim = p_sim
                    if p_sim < self.max_similarity:
                        break
            else:
                warnings.warn(
                    'Could not create a semantic pointer with '
                    'max_similarity=%1.2f (D=%d, M=%d)'
                    % (self.max_similarity, self.dimensions,
                       len(self.pointers)))
        return best_p

    def __contains__(self, key):
        return key in self.pointers

    def __len__(self):
        return len(self._vectors)

    def __iter__(self):
        return iter(self._keys)

    def __getitem__(self, key):
        """Return the semantic pointer with the requested name."""
        if not self.strict and key not in self:
            self.add(key, self.create_pointer())
        return self.pointers[key]

    def __hash__(self):
        return hash(id(self))

    def add(self, key, p):
        """Add a new semantic pointer to the vocabulary.

        The pointer value can be a `.SemanticPointer` or a vector.
        """
        if not valid_sp_regex.match(key):
            raise SpaParseError(
                "Invalid Semantic Pointer name. Valid names are valid "
                "Python 2 identifiers beginning with a capital letter.")
        if not isinstance(p, pointer.SemanticPointer):
            p = pointer.SemanticPointer(p)

        if key in self.pointers:
            raise ValidationError("The semantic pointer %r already exists"
                                  % key, attr='pointers', obj=self)

        self.pointers[key] = p
        self._keys.append(key)
        self._vectors = np.vstack([self._vectors, p.v])

    def populate(self, pointers):
        for p_expr in pointers.split(';'):
            assign_split = p_expr.split('=', 1)
            modifier_split = p_expr.split('.', 1)
            if len(assign_split) > 1:
                name, value_expr = assign_split
                value = eval(value_expr.strip(), {}, self)
            elif len(modifier_split) > 1:
                name = modifier_split[0]
                value = self.create_pointer(transform=modifier_split[1])
            else:
                name = p_expr
                value = self.create_pointer()
            self.add(name.strip(), value)

    def parse(self, text):
        """Evaluate a text string and return the corresponding SemanticPointer.

        This uses the Python ``eval()`` function, so any Python operators that
        have been defined for SemanticPointers are valid (``+``, ``-``, ``*``,
        ``~``, ``()``). Valid semantic pointer terms must start
        with a capital letter.

        If the expression returns a scalar (int or float), a scaled version
        of the identity SemanticPointer will be returned.
        """

        # The following line does everything.  Note that self is being
        # passed in as the locals dictionary, and thanks to the __getitem__
        # implementation, this will automatically create new semantic
        # pointers as needed.
        try:
            value = eval(text, {}, self)
        except NameError:
            raise SpaParseError(
                "Semantic pointers must start with a capital letter.")

        if is_number(value):
            value *= Identity(self.dimensions)
        elif not isinstance(value, pointer.SemanticPointer):
            raise SpaParseError(
                "The result of parsing '%s' is not a SemanticPointer." % text)
        return value

    def parse_n(self, *texts):
        return [self.parse(t) for t in texts]

    def dot(self, v):
        """Returns the dot product with all terms in the Vocabulary.

        Input parameter can either be a `.SemanticPointer` or a vector.
        """
        if isinstance(v, pointer.SemanticPointer):
            v = v.v
        return np.dot(self._vectors, v)

    def transform_to(self, other, populate=None, keys=None):
        """Create a linear transform from one Vocabulary to another.

        This is simply the sum of the outer products of the corresponding
        terms in each Vocabulary.

        Parameters
        ----------
        other : Vocabulary
            The other vocabulary to translate into.
        keys : list, optional (Default: None)
            If None, any term that exists in just one of the Vocabularies
            will be created in the other Vocabulary and included. Otherwise,
            the transformation will only consider terms in this list. Any
            terms in this list that do not exist in the Vocabularies will
            be created.
        """
        if keys is None:
            keys = self._keys

        t = np.zeros((other.dimensions, self.dimensions), dtype=float)
        for k in keys:
            if k not in other:
                if populate is None:
                    warnings.warn(NengoWarning(
                        "The transform_to source vocabulary has keys not "
                        "existent in the target vocabulary. These will "
                        "be ignored. Use the `populate=False` keyword "
                        "argument to silence this warning or "
                        "`populate=True` to automatically add missing "
                        "keys to the target vocabulary."))
                elif populate:
                    other.populate(k)
                else:
                    continue
            a = self[k].v
            b = other[k].v
            t += np.outer(b, a)
        return t

    def create_subset(self, keys):
        """Returns the subset of this vocabulary.

        Creates and returns a subset of the current vocabulary that contains
        all the semantic pointers found in keys.

        Parameters
        ----------
        keys : list
            List of semantic pointer names to be copied over to the
            new vocabulary.
        """
        # Make new Vocabulary object
        subset = Vocabulary(self.dimensions, self.strict, self.max_similarity,
                            self.rng)

        # Copy over the new keys
        for key in keys:
            subset.add(key, self.pointers[key])

        return subset


class VocabularyMap(Mapping):
    """Maps dimensionalities to corresponding vocabularies."""
    def __init__(self, vocabs=None, rng=None):
        if vocabs is None:
            vocabs = []
        self.rng = rng

        self._vocabs = {}
        try:
            for vo in vocabs:
                self.add(vo)
        except (AttributeError, TypeError):
            raise ValueError(
                "The `vocabs` argument requires a list of Vocabulary "
                "instances or `None`.")

    def add(self, vocab):
        if vocab.dimensions in self._vocabs:
            warnings.warn("Duplicate vocabularies with dimension %d. "
                          "Using the last entry in the vocab list with "
                          "that dimensionality." % (vocab.dimensions))
        self._vocabs[vocab.dimensions] = vocab

    def __delitem__(self, dimensions):
        del self._vocabs[dimensions]

    def discard(self, vocab):
        if isinstance(vocab, int):
            del self._vocabs[vocab]
        elif self._vocabs.get(vocab.dimensions, None) is vocab:
            del self._vocabs[vocab.dimensions]

    def __getitem__(self, dimensions):
        return self._vocabs[dimensions]

    def get_or_create(self, dimensions):
        if dimensions not in self._vocabs:
            self._vocabs[dimensions] = Vocabulary(
                dimensions, strict=False, rng=self.rng)
        return self._vocabs[dimensions]

    def __iter__(self):
        return iter(self._vocabs)

    def __len__(self):
        return len(self._vocabs)

    def __contains__(self, vocab):
        if isinstance(vocab, int):
            return vocab in self._vocabs
        else:
            return (vocab.dimensions in self._vocabs and
                    self._vocabs[vocab.dimensions] is vocab)


class VocabularyMapParam(nengo.params.Parameter):
    """Can be a mapping from dimensions to vocabularies."""

    def validate(self, instance, vocab_set):
        super(VocabularyMapParam, self).validate(instance, vocab_set)

        if vocab_set is not None and not isinstance(vocab_set, VocabularyMap):
            try:
                VocabularyMap(vocab_set)
            except ValueError:
                raise ValidationError(
                    "Must be of type 'VocabularyMap' or compatible "
                    "(got type %r)."
                    % type(vocab_set).__name__, attr=self.name, obj=instance)

        return vocab_set

    def __set__(self, instance, value):
        if not isinstance(value, VocabularyMap):
            value = VocabularyMap(value)
        super(VocabularyMapParam, self).__set__(instance, value)


class VocabularyOrDimParam(nengo.params.Parameter):
    """Can be a vocabulary or integer denoting a dimensionality."""

    def validate(self, instance, value):
        super(VocabularyOrDimParam, self).validate(instance, value)

        if value is not None:
            if is_integer(value):
                if value < 1:
                    raise ValidationError(
                        "Vocabulary dimensionality must be at least 1.",
                        attr=self.name, obj=instance)
            elif not isinstance(value, Vocabulary):
                raise ValidationError(
                    "Must be of type 'Vocabulary' or an integer (got type %r)."
                    % type(value).__name__, attr=self.name, obj=instance)

    def __set__(self, instance, value):
        if is_integer(value):
            value = instance.vocabs.get_or_create(value)
        super(VocabularyOrDimParam, self).__set__(instance, value)



import pytest

from nengo_spa.math import prob_cleanup


def test_prob_cleanup(rng):
    pytest.importorskip('scipy')

    assert 1.0 > prob_cleanup(0.7, 64, 10000) > 0.9999
    assert 0.9999 > prob_cleanup(0.6, 64, 10000) > 0.999
    assert 0.99 > prob_cleanup(0.5, 64, 1000) > 0.9

    assert 0.999 > prob_cleanup(0.4, 128, 1000) > 0.997
    assert 0.99 > prob_cleanup(0.4, 128, 10000) > 0.97
    assert 0.9 > prob_cleanup(0.4, 128, 100000) > 0.8



import nengo
from nengo.params import Default, IntParam
from nengo_spa.network import Network


class Product(Network):
    """A SPA network capable of multiplying two scalars.

    Parameters
    ----------
    n_neurons : int, optional (Default: 200)
        Number of neurons to use in product computation.
    kwargs
        Keyword arguments passed through to ``spa.Network``.
    """

    n_neurons = IntParam('n_neurons', default=200, low=1, readonly=True)

    def __init__(self, n_neurons=Default, **kwargs):
        super(Product, self).__init__(**kwargs)

        self.n_neurons = n_neurons

        with self:
            self.product = nengo.networks.Product(self.n_neurons, 1)

        self.input_a = self.product.input_a
        self.input_b = self.product.input_b
        self.output = self.product.output
        self.inputs = dict(
            input_a=(self.input_a, None),
            input_b=(self.input_b, None))
        self.outputs = dict(default=(self.output, None))



import nengo
import nengo_spa as spa


def test_basic():
    with spa.Network() as model:
        model.compare = spa.Compare(vocab=16)

    inputA = model.get_network_input('compare.input_a')
    inputB = model.get_network_input('compare.input_b')
    output = model.get_network_output('compare')
    # all nodes should be acquired correctly
    assert inputA[0] is model.compare.input_a
    assert inputB[0] is model.compare.input_b
    assert output[0] is model.compare.output
    # all inputs should share the same vocab
    assert inputA[1] is inputB[1]
    assert inputA[1].dimensions == 16
    # output should have no vocab
    assert output[1] is None


def test_run(Simulator, seed):
    with spa.Network(seed=seed) as model:
        model.compare = spa.Compare(vocab=16)
        model.compare.vocab.populate('A; B')

        def inputA(t):
            if 0 <= t < 0.1:
                return 'A'
            else:
                return 'B'

        model.input = spa.Input()
        model.input.compare.input_a = inputA
        model.input.compare.input_b = 'A'

    compare, vocab = model.get_network_output('compare')

    with model:
        p = nengo.Probe(compare, 'output', synapse=0.03)

    with Simulator(model) as sim:
        sim.run(0.2)

    assert sim.data[p][100] > 0.8
    assert sim.data[p][199] < 0.2



import numpy as np
import pytest

import nengo
import nengo_spa as spa
from nengo_spa.exceptions import SpaNetworkError, SpaTypeError
from nengo_spa.examine import similarity
from nengo_spa.vocab import VocabularyMap


class SpaCommunicationChannel(spa.Network):
    def __init__(
            self, dimensions, label=None, seed=None, add_to_container=None):
        super(SpaCommunicationChannel, self).__init__(
            label, seed, add_to_container)

        with self:
            self.state_in = spa.State(dimensions)
            self.state_out = spa.State(dimensions)

            spa.Actions('state_out = state_in').build()

        self.inputs = dict(
            default=self.state_in.inputs['default'],
            secondary=self.state_in.inputs['default'])
        self.outputs = dict(
            default=self.state_out.outputs['default'],
            secondary=self.state_out.outputs['default'])


def test_spa_verification(seed, plt):
    d = 16

    model = spa.Network(seed=seed)

    # building a normal model that shouldn't raise a warning
    with model:
        model.buf = spa.State(d)
        model.input_node = spa.Input()
        model.input_node.buf = 'B'
        # make sure errors aren't fired for non-spa networks
        prod = nengo.networks.Product(10, 2)  # noqa: F841
        model.int_val = 1

        # reassignment is fine for non-networks
        model.int_val = 2

    # reassignment of networks should throw an error
    with pytest.raises(ValueError):
        with model:
            model.buf = spa.State(d, feedback=1)


def test_spa_network_exception():
    class MyException(Exception):
        pass

    class TestNetwork(spa.network.Network):
        def __init__(self):
            super(TestNetwork, self).__init__()
            raise MyException()

    with pytest.raises(MyException):
        with spa.Network() as model:
            model.test = TestNetwork()


def test_spa_get():
    D = 16
    model = spa.Network()
    with model:
        model.buf1 = spa.State(D)
        model.buf2 = spa.State(D)
        model.compare = spa.Compare(D)

    assert model.get_spa_network('buf1') is model.buf1
    with pytest.raises(SpaNetworkError):
        model.get_spa_network('buf1.default')
    assert model.get_spa_network('buf2') is model.buf2
    assert model.get_network_input('buf1')[0] is model.buf1.input
    assert model.get_network_input('buf1.default')[0] is model.buf1.input
    assert model.get_network_output('buf1')[0] is model.buf1.output
    assert model.get_network_output('buf1.default')[0] is model.buf1.output
    assert model.get_network_input(
        'compare.input_a')[0] is model.compare.input_a
    assert model.get_network_input(
        'compare.input_b')[0] is model.compare.input_b

    with pytest.raises(SpaNetworkError):
        model.get_spa_network('dummy')
    with pytest.raises(SpaNetworkError):
        model.get_network_input('dummy')
    with pytest.raises(SpaNetworkError):
        model.get_network_output('dummy')
    with pytest.raises(SpaNetworkError):
        model.get_network_input('buf1.A')
    with pytest.raises(SpaNetworkError):
        model.get_network_input('compare')


def test_spa_vocab():
    # create a model without a vocab and check that it is empty
    model = spa.Network()
    assert len(model.vocabs) == 0

    # create a model with a vocab and check that it's filled
    va = spa.Vocabulary(16)
    va.populate("PANTS")
    vb = spa.Vocabulary(32)
    vb.populate("SHOES")
    model = spa.Network(vocabs=VocabularyMap([va, vb]))
    assert list(model.vocabs[16].keys()) == ["PANTS"]
    assert list(model.vocabs[32].keys()) == ["SHOES"]

    # warning on vocabs with duplicate dimensions
    vc = spa.Vocabulary(16)
    vc.populate("SOCKS")
    with pytest.warns(UserWarning):
        model = spa.Network(vocabs=VocabularyMap([va, vb, vc]))
    assert list(model.vocabs[16].keys()) == ["SOCKS"]
    assert list(model.vocabs[32].keys()) == ["SHOES"]


def test_hierarchical(Simulator, seed, plt):
    d = 32

    with spa.Network(seed=seed) as model:
        model.comm_channel = SpaCommunicationChannel(d)
        model.out = spa.State(d)

        spa.Actions('out = comm_channel').build()
        model.stimulus = spa.Input()
        model.stimulus.comm_channel = 'A'

        p = nengo.Probe(model.out.output, synapse=0.03)

    with Simulator(model) as sim:
        sim.run(0.5)

    t = sim.trange() > 0.2
    v = model.vocabs[d].parse('A').v

    plt.plot(sim.trange(), similarity(sim.data[p], v))
    plt.xlabel("t [s]")
    plt.ylabel("Similarity")

    assert np.mean(similarity(sim.data[p][t], v)) > 0.8


def test_hierarichal_network_name_resolution():
    with spa.Network() as model:
        model.comm_channel = SpaCommunicationChannel(16)

    assert (
        model.get_spa_network('comm_channel.state_in') is
        model.comm_channel.state_in)
    assert (
        model.get_network_input('comm_channel.state_in') is
        model.comm_channel.state_in.inputs['default'])
    assert (
        model.get_network_input('comm_channel.secondary') is
        model.comm_channel.inputs['secondary'])
    assert (
        model.get_network_output('comm_channel.state_out') is
        model.comm_channel.state_out.outputs['default'])
    assert (
        model.get_network_output('comm_channel.secondary') is
        model.comm_channel.outputs['secondary'])


def test_hierarchical_actions(Simulator, seed, plt):
    d = 32

    with spa.Network(seed=seed) as model:
        model.comm_channel = SpaCommunicationChannel(d)
        model.out = spa.State(d)

        spa.Actions('out = comm_channel.state_out').build()
        model.stimulus = spa.Input()
        model.stimulus.comm_channel.state_in = 'A'

        p = nengo.Probe(model.out.output, synapse=0.03)

    with Simulator(model) as sim:
        sim.run(0.5)

    t = sim.trange() > 0.2
    v = model.vocabs[d].parse('A').v

    plt.plot(sim.trange(), similarity(sim.data[p], v))
    plt.xlabel("t [s]")
    plt.ylabel("Similarity")

    assert np.mean(similarity(sim.data[p][t], v)) > 0.8


def test_vocab_config():
    with spa.Network() as model:
        with spa.Network() as model.shared_vocabs:
            pass
        with spa.Network(
                vocabs=spa.VocabularyMap()) as model.non_shared_vocabs:
            pass

    assert model.shared_vocabs.vocabs is model.vocabs
    assert model.non_shared_vocabs.vocabs is not model.vocabs


def test_no_magic_vocab_transform():
    d = 16
    v1 = spa.Vocabulary(d)
    v2 = spa.Vocabulary(d)

    with spa.Network() as model:
        model.a = spa.State(vocab=v1)
        model.b = spa.State(vocab=v2)
        with pytest.raises(SpaTypeError):
            spa.Actions('b = a').build()


@pytest.mark.parametrize('d1,d2,method,lookup', [
    (16, 16, 'reinterpret(a, v2)', 'v1'),
    (16, 16, 'reinterpret(a)', 'v1'),
    (16, 16, 'reinterpret(a, b)', 'v1'),
    (16, 32, 'translate(a, v2)', 'v2'),
    (16, 32, 'translate(a)', 'v2'),
    (16, 32, 'translate(a, b)', 'v2')])
def test_casting_vocabs(d1, d2, method, lookup, Simulator, plt, rng):
    v1 = spa.Vocabulary(d1, rng=rng)
    v1.populate('A')
    v2 = spa.Vocabulary(d2, rng=rng)
    v2.populate('A')

    with spa.Network() as model:
        model.a = spa.State(vocab=v1)
        model.b = spa.State(vocab=v2)
        spa.Actions('b = {}'.format(method), vocabs={'v2': v2}).build()
        model.stimulus = spa.Input()
        model.stimulus.a = 'A'
        p = nengo.Probe(model.b.output, synapse=0.03)

    with Simulator(model) as sim:
        sim.run(0.5)

    t = sim.trange() > 0.2
    v = locals()[lookup].parse('A').v

    plt.plot(sim.trange(), similarity(sim.data[p], v))
    plt.xlabel("t [s]")
    plt.ylabel("Similarity")

    assert np.mean(similarity(sim.data[p][t], v)) > 0.8


def test_copy_spa(RefSimulator):
    with spa.Network() as original:
        original.state = spa.State(16)
        spa.Actions("state = A").build()

    cp = original.copy()

    # Check that it still builds.
    with RefSimulator(cp):
        pass



import numpy as np

import nengo
import nengo_spa as spa
from nengo_spa.vocab import VocabularyMap
from nengo.utils.numpy import rmse


def test_basic():
    with spa.Network() as model:
        model.bind = spa.Bind(vocab=16)

    inputA = model.get_network_input('bind.input_a')
    inputB = model.get_network_input('bind.input_b')
    output = model.get_network_output('bind')
    # all nodes should be acquired correctly
    assert inputA[0] is model.bind.input_a
    assert inputB[0] is model.bind.input_b
    assert output[0] is model.bind.output
    # all inputs and outputs should share the same vocab
    assert inputA[1] is inputB[1]
    assert inputA[1].dimensions == 16
    assert output[1].dimensions == 16


def test_run(Simulator, seed):
    rng = np.random.RandomState(seed)
    vocab = spa.Vocabulary(32, rng=rng)
    vocab.populate('A; B')

    with spa.Network(seed=seed, vocabs=VocabularyMap([vocab])) as model:
        model.bind = spa.Bind(vocab=32)

        def inputA(t):
            if 0 <= t < 0.1:
                return 'A'
            else:
                return 'B'

        model.input = spa.Input()
        model.input.bind.input_a = inputA
        model.input.bind.input_b = 'A'

    bind, vocab = model.get_network_output('bind')

    with model:
        p = nengo.Probe(bind, 'output', synapse=0.03)

    with Simulator(model) as sim:
        sim.run(0.2)

    error = rmse(vocab.parse("(B*A).normalized()").v, sim.data[p][-1])
    assert error < 0.1

    error = rmse(vocab.parse("(A*A).normalized()").v, sim.data[p][100])
    assert error < 0.1



from nengo_spa import version

from nengo_spa.actions import Actions
from nengo_spa.examine import pairs, similarity, text
from nengo_spa.modules import (
    AssociativeMemory, IAAssocMem, ThresholdingAssocMem, WTAAssocMem,
    BasalGanglia,
    Bind,
    Compare,
    Input,
    Product,
    Scalar,
    State,
    Thalamus)
from nengo_spa.network import Network
from nengo_spa.vocab import Vocabulary, VocabularyMap


__copyright__ = "2013-2017, Applied Brain Research"
__license__ = "Free for non-commercial use; see LICENSE.rst"
__version__ = version.version



"""Associative memory implementations.

See :doc:`examples/associative_memory` for an introduction and examples.
"""
import nengo
import numpy as np
from nengo_spa.network import Network
from nengo_spa.networks.selection import IA, Thresholding, WTA
from nengo_spa.vocab import VocabularyOrDimParam
from nengo.utils.network import with_self


class AssociativeMemory(Network):
    """General associative memory network.

    This provides a low-level selection network with the necessary interface
    to include it within the SPA system.

    Parameters
    ----------
    selection_net : Network
        The network that is used to select the response. It needs to accept
        the arguments *n_neurons* (number of neurons to use to represent each
        possible choice) and *n_ensembles* (number of choices). The returned
        network needs to have an *input* attribute to which the utilities for
        each choice are connected and an *output* attribute from which a
        connection will be created to read the selected output(s).
    input_vocab: Vocabulary
        The vocabulary to match.
    output_vocab: Vocabulary, optional (Default: None)
        The vocabulary to be produced for each match. If
        None, the associative memory will act like an autoassociative memory
        (cleanup memory).
    input_keys : list, optional (Default: None)
        A list of strings that correspond to the input vectors.
    output_keys : list, optional (Default: None)
        A list of strings that correspond to the output vectors.
    n_neurons : int
        Number of neurons to represent each choice, passed on to the
        *selection_net*.
    label : str, optional (Default: None)
        A name for the ensemble. Used for debugging and visualization.
    seed : int, optional (Default: None)
        The seed used for random number generation.
    add_to_container : bool, optional (Default: None)
        Determines if this Network will be added to the current container.
        If None, will be true if currently with
    vocabs : VocabularyMap, optional (Default: None)
        Maps dimensionalities to the corresponding default vocabularies.
    """
    input_vocab = VocabularyOrDimParam(
        'input_vocab', default=None, readonly=True)
    output_vocab = VocabularyOrDimParam(
        'output_vocab', default=None, readonly=True)

    def __init__(
            self, selection_net, input_vocab, output_vocab=None,
            input_keys=None, output_keys=None, n_neurons=50,
            label=None, seed=None, add_to_container=None, vocabs=None,
            **selection_net_args):
        super(AssociativeMemory, self).__init__(
            label=label, seed=seed, add_to_container=add_to_container,
            vocabs=vocabs)

        if output_vocab is None:
            output_vocab = input_vocab
        self.input_vocab = input_vocab
        self.output_vocab = output_vocab

        if input_keys is None:
            input_keys = self.input_vocab.keys()
            input_vectors = self.input_vocab.vectors
        else:
            input_vectors = [input_vocab.parse(key).v for key in input_keys]

        if output_keys is None:
            output_keys = input_keys
        output_vectors = [output_vocab.parse(key).v for key in output_keys]

        input_vectors = np.asarray(input_vectors)
        output_vectors = np.asarray(output_vectors)

        with self:
            self.selection = selection_net(
                n_neurons=n_neurons, n_ensembles=len(input_vectors),
                label="selection", **selection_net_args)
            self.input = nengo.Node(size_in=self.input_vocab.dimensions,
                                    label="input")
            self.output = nengo.Node(size_in=self.output_vocab.dimensions,
                                     label="output")

            nengo.Connection(
                self.input, self.selection.input, transform=input_vectors)
            nengo.Connection(
                self.selection.output, self.output, transform=output_vectors.T)

        self.inputs = dict(default=(self.input, self.input_vocab))
        self.outputs = dict(default=(self.output, self.output_vocab))

    @with_self
    def add_default_output(self, key, min_activation_value, n_neurons=50):
        """Adds a Semantic Pointer to output when no other pointer is active.

        Parameters
        ----------
        key : str
            Semantic Pointer to output.
        min_activation_value : float
            Minimum output of another Semantic Pointer to deactivate the
            default output.
        n_neurons : int, optional (Default: 50)
            Number of neurons used to represent the default Semantic Pointer.
        """
        assert not hasattr(self, 'default_ens'), \
            "Can add default output only once."

        with nengo.presets.ThresholdingEnsembles(0.):
            setattr(self, 'default_ens',
                    nengo.Ensemble(n_neurons, 1, label="default"))
        setattr(self, 'bias', nengo.Node(1., label="bias"))
        nengo.Connection(self.bias, self.default_ens)
        nengo.Connection(
            self.default_ens, self.output,
            transform=np.atleast_2d(self.output_vocab.parse(key).v).T)
        nengo.Connection(
            self.selection.output, self.default_ens,
            transform=-np.ones(
                (1, self.selection.output.size_out)) / min_activation_value)


class IAAssocMem(AssociativeMemory):
    """Associative memory based on the `IA` network.

    See `AssociativeMemory` and `IA` for more information.
    """
    def __init__(
            self, input_vocab, output_vocab=None, input_keys=None,
            output_keys=None, n_neurons=50, label=None, seed=None,
            add_to_container=None, vocabs=None, **selection_net_args):
        super(IAAssocMem, self).__init__(
            selection_net=IA,
            input_vocab=input_vocab, output_vocab=output_vocab,
            input_keys=input_keys, output_keys=output_keys,
            n_neurons=n_neurons, label=label, seed=seed,
            add_to_container=add_to_container, vocabs=vocabs,
            **selection_net_args)
        self.input_reset = self.selection.input_reset
        self.inputs['reset'] = (self.input_reset, None)


class ThresholdingAssocMem(AssociativeMemory):
    """Associative memory based on `Thresholding`.

    See `AssociativeMemory` and `Thresholding` for more information.
    """
    def __init__(
            self, threshold, input_vocab, output_vocab=None, input_keys=None,
            output_keys=None, n_neurons=50, label=None, seed=None,
            add_to_container=None, vocabs=None, **selection_net_args):
        selection_net_args['threshold'] = threshold
        super(ThresholdingAssocMem, self).__init__(
            selection_net=Thresholding,
            input_vocab=input_vocab, output_vocab=output_vocab,
            input_keys=input_keys, output_keys=output_keys,
            n_neurons=n_neurons, label=label, seed=seed,
            add_to_container=add_to_container, vocabs=vocabs,
            **selection_net_args)


class WTAAssocMem(AssociativeMemory):
    """Associative memory based on the `WTA` network.

    See `AssociativeMemory` and `WTA` for more information.
    """
    def __init__(
            self, threshold, input_vocab, output_vocab=None, input_keys=None,
            output_keys=None, n_neurons=50, label=None, seed=None,
            add_to_container=None, vocabs=None, **selection_net_args):
        selection_net_args['threshold'] = threshold
        super(WTAAssocMem, self).__init__(
            selection_net=WTA,
            input_vocab=input_vocab, output_vocab=output_vocab,
            input_keys=input_keys, output_keys=output_keys,
            n_neurons=n_neurons, label=label, seed=seed,
            add_to_container=add_to_container, vocabs=vocabs,
            **selection_net_args)



import nengo
import numpy as np
import nengo_spa as spa
from nengo_spa import Vocabulary
from nengo_spa.modules.assoc_mem import ThresholdingAssocMem, WTAAssocMem
from nengo_spa.examine import similarity


filtered_step_fn = lambda x: np.maximum(1. - np.exp(-15. * x), 0.)


def test_am_basic(Simulator, plt, seed, rng):
    """Basic associative memory test."""

    d = 64
    vocab = Vocabulary(d, rng=rng)
    vocab.populate('A; B; C; D')

    with spa.Network('model', seed=seed) as m:
        m.am = ThresholdingAssocMem(threshold=0.3, input_vocab=vocab,
                                    function=filtered_step_fn)
        m.stimulus = spa.Input()
        m.stimulus.am = 'A'

        in_p = nengo.Probe(m.am.input)
        out_p = nengo.Probe(m.am.output, synapse=0.03)

    with Simulator(m) as sim:
        sim.run(0.2)
    t = sim.trange()

    plt.subplot(3, 1, 1)
    plt.plot(t, similarity(sim.data[in_p], vocab))
    plt.ylabel("Input")
    plt.ylim(top=1.1)
    plt.subplot(3, 1, 2)
    plt.plot(t, similarity(sim.data[out_p], vocab))
    plt.plot(t[t > 0.15], np.ones(t.shape)[t > 0.15] * 0.95, c='g', lw=2)
    plt.ylabel("Output")

    assert np.all(similarity(sim.data[in_p][t > 0.15], vocab)[:, 0] > 0.99)
    assert np.all(similarity(sim.data[out_p][t > 0.15], vocab)[:, 0] > 0.95)
    assert np.all(similarity(sim.data[out_p][t > 0.15], vocab)[:, 1:] < 0.1)


def test_am_threshold(Simulator, plt, seed, rng):
    """Associative memory thresholding with differing input/output vocabs."""
    d = 64
    vocab = Vocabulary(d, rng=rng)
    vocab.populate('A; B; C; D')

    d2 = int(d / 2)
    vocab2 = Vocabulary(d2, rng=rng)
    vocab2.populate('A; B; C; D')

    def input_func(t):
        return '0.49 * A' if t < 0.1 else '0.8 * B'

    with spa.Network('model', seed=seed) as m:
        m.am = ThresholdingAssocMem(
            threshold=0.5, input_vocab=vocab, output_vocab=vocab2,
            function=filtered_step_fn)
        m.stimulus = spa.Input()
        m.stimulus.am = input_func

        in_p = nengo.Probe(m.am.input)
        out_p = nengo.Probe(m.am.output, synapse=0.03)

    with Simulator(m) as sim:
        sim.run(0.3)
    t = sim.trange()
    below_th = t < 0.1
    above_th = t > 0.25

    plt.subplot(2, 1, 1)
    plt.plot(t, similarity(sim.data[in_p], vocab))
    plt.ylabel("Input")
    plt.subplot(2, 1, 2)
    plt.plot(t, similarity(sim.data[out_p], vocab2))
    plt.plot(t[above_th], np.ones(t.shape)[above_th] * 0.9, c='g', lw=2)
    plt.ylabel("Output")

    assert np.mean(sim.data[out_p][below_th]) < 0.01
    assert np.all(
        similarity(sim.data[out_p][above_th], [vocab2['B'].v]) > 0.90)


def test_am_wta(Simulator, plt, seed, rng):
    """Test the winner-take-all ability of the associative memory."""

    d = 64
    vocab = Vocabulary(d, rng=rng)
    vocab.populate('A; B; C; D')

    def input_func(t):
        if t < 0.2:
            return 'A + 0.8 * B'
        elif t < 0.3:
            return '0'
        else:
            return '0.8 * A + B'

    with spa.Network('model', seed=seed) as m:
        m.am = WTAAssocMem(
            threshold=0.3, input_vocab=vocab, function=filtered_step_fn)
        m.stimulus = spa.Input()
        m.stimulus.am = input_func

        in_p = nengo.Probe(m.am.input)
        out_p = nengo.Probe(m.am.output, synapse=0.03)

    with Simulator(m) as sim:
        sim.run(0.5)
    t = sim.trange()
    more_a = (t > 0.15) & (t < 0.2)
    more_b = t > 0.45

    plt.subplot(2, 1, 1)
    plt.plot(t, similarity(sim.data[in_p], vocab))
    plt.ylabel("Input")
    plt.ylim(top=1.1)
    plt.subplot(2, 1, 2)
    plt.plot(t, similarity(sim.data[out_p], vocab))
    plt.plot(t[more_a], np.ones(t.shape)[more_a] * 0.9, c='g', lw=2)
    plt.plot(t[more_b], np.ones(t.shape)[more_b] * 0.9, c='g', lw=2)
    plt.ylabel("Output")

    assert np.all(similarity(sim.data[out_p][more_a], [vocab['A'].v]) > 0.9)
    assert np.all(similarity(sim.data[out_p][more_a], [vocab['B'].v]) < 0.1)
    assert np.all(similarity(sim.data[out_p][more_b], [vocab['B'].v]) > 0.79)
    assert np.all(similarity(sim.data[out_p][more_b], [vocab['A'].v]) < 0.1)


def test_am_default_output(Simulator, plt, seed, rng):
    d = 64
    vocab = Vocabulary(d, rng=rng)
    vocab.populate('A; B; C; D')

    def input_func(t):
        return '0.2 * A' if t < 0.25 else 'A'

    with spa.Network('model', seed=seed) as m:
        m.am = ThresholdingAssocMem(threshold=0.5, input_vocab=vocab,
                                    function=filtered_step_fn)
        m.am.add_default_output('D', 0.5)
        m.stimulus = spa.Input()
        m.stimulus.am = input_func

        in_p = nengo.Probe(m.am.input)
        out_p = nengo.Probe(m.am.output, synapse=0.03)

    with Simulator(m) as sim:
        sim.run(0.5)
    t = sim.trange()
    below_th = (t > 0.15) & (t < 0.25)
    above_th = t > 0.4

    plt.subplot(2, 1, 1)
    plt.plot(t, similarity(sim.data[in_p], vocab))
    plt.ylabel("Input")
    plt.subplot(2, 1, 2)
    plt.plot(t, similarity(sim.data[out_p], vocab))
    plt.plot(t[below_th], np.ones(t.shape)[below_th] * 0.9, c='c', lw=2)
    plt.plot(t[above_th], np.ones(t.shape)[above_th] * 0.9, c='b', lw=2)
    plt.plot(t[above_th], np.ones(t.shape)[above_th] * 0.1, c='c', lw=2)
    plt.ylabel("Output")

    assert np.all(similarity(sim.data[out_p][below_th], [vocab['D'].v]) > 0.9)
    assert np.all(similarity(sim.data[out_p][above_th], [vocab['D'].v]) < 0.15)
    assert np.all(similarity(sim.data[out_p][above_th], [vocab['A'].v]) > 0.9)


def test_am_spa_keys_as_expressions(Simulator, plt, seed, rng):
    """Provide semantic pointer expressions as input and output keys."""
    d = 64

    vocab_in = Vocabulary(d, rng=rng)
    vocab_out = Vocabulary(d, rng=rng)

    vocab_in.populate('A; B')
    vocab_out.populate('C; D')

    in_keys = ['A', 'A*B']
    out_keys = ['C*D', 'C+D']

    with spa.Network(seed=seed) as model:
        model.am = ThresholdingAssocMem(
            threshold=0.3, input_vocab=vocab_in, output_vocab=vocab_out,
            input_keys=in_keys, output_keys=out_keys)

        model.inp = spa.Input()
        model.inp.am = lambda t: 'A' if t < 0.1 else 'A*B'

        in_p = nengo.Probe(model.am.input)
        out_p = nengo.Probe(model.am.output, synapse=0.03)

    with nengo.Simulator(model) as sim:
        sim.run(0.2)

    # Specify t ranges
    t = sim.trange()
    t_item1 = (t > 0.075) & (t < 0.1)
    t_item2 = (t > 0.175) & (t < 0.2)

    # Modify vocabularies (for plotting purposes)
    vocab_in.add(in_keys[1], vocab_in.parse(in_keys[1]).v)
    vocab_out.add(out_keys[0], vocab_out.parse(out_keys[0]).v)

    plt.subplot(2, 1, 1)
    plt.plot(t, similarity(sim.data[in_p], vocab_in))
    plt.ylabel("Input: " + ', '.join(in_keys))
    plt.legend(vocab_in.keys, loc='best')
    plt.ylim(top=1.1)
    plt.subplot(2, 1, 2)
    for t_item, c, k in zip([t_item1, t_item2], ['b', 'g'], out_keys):
        plt.plot(t, similarity(
            sim.data[out_p], [vocab_out.parse(k).v], normalize=True),
            label=k, c=c)
        plt.plot(t[t_item], np.ones(t.shape)[t_item] * 0.9, c=c, lw=2)
    plt.ylabel("Output: " + ', '.join(out_keys))
    plt.legend(loc='best')

    assert np.mean(similarity(sim.data[out_p][t_item1],
                              vocab_out.parse(out_keys[0]).v,
                              normalize=True)) > 0.9
    assert np.mean(similarity(sim.data[out_p][t_item2],
                              vocab_out.parse(out_keys[1]).v,
                              normalize=True)) > 0.9
